##
#  File:           PharosTargetActivityProvider.py
#  Date:           17-Jun-2021 jdw
#
#  Updated:
#  3-Mar-2023 aae fix method typo
#
##
"""
Accessors for Pharos target activity data.

"""

import datetime
import logging
import os.path
import time

from rcsb.utils.chemref.PharosProvider import PharosProvider
from rcsb.utils.io.FileUtil import FileUtil
from rcsb.utils.io.MarshalUtil import MarshalUtil
from rcsb.utils.io.StashableBase import StashableBase


logger = logging.getLogger(__name__)


class PharosTargetActivityProvider(StashableBase):
    """Accessors for Pharos target activity data."""

    def __init__(self, cachePath, useCache):
        #
        self.__cachePath = cachePath
        self.__dirName = "Pharos-target-activity"
        super(PharosTargetActivityProvider, self).__init__(self.__cachePath, [self.__dirName])
        self.__dirPath = os.path.join(self.__cachePath, self.__dirName)
        self.__mU = MarshalUtil(workPath=self.__cachePath)
        #
        self.__aD, self.__tD, self.__version = self.__reload(self.__dirPath, useCache)

    def testCache(self, minCount=1):
        if self.__aD and (len(self.__aD) >= minCount):
            logger.info("Cached Pharos activity data for (%d) targets", len(self.__aD))
            return True
        return False

    def getAssignmentVersion(self):
        return self.__version

    def getTargetActivityDataPath(self):
        return os.path.join(self.__dirPath, "pharos-target-activity-data.json")

    def reload(self):
        self.__aD, self.__tD, self.__version = self.__reload(self.__dirPath, useCache=True)
        return True

    def __reload(self, dirPath, useCache):
        startTime = time.time()
        aD = {}
        tD = {}
        version = None
        fU = FileUtil()
        fU.mkdir(dirPath)
        targetActivityFilePath = self.getTargetActivityDataPath()
        #
        if useCache and fU.exists(targetActivityFilePath):
            logger.info("useCache %r using %r", useCache, targetActivityFilePath)
            qD = self.__mU.doImport(targetActivityFilePath, fmt="json")
            aD = qD["activity"] if "activity" in qD else {}
            tD = qD["targets"] if "targets" in qD else {}
            version = qD["version"] if "version" in qD else None
        #
        logger.info("Completed reload of (%d) activity records at %s (%.4f seconds)", len(aD), time.strftime("%Y %m %d %H:%M:%S", time.localtime()), time.time() - startTime)
        #
        return aD, tD, version

    def getTargetActivity(self, pharosTargetId):
        try:
            return self.__aD[pharosTargetId] if pharosTargetId in self.__aD else []
        except Exception:
            return []

    def hasTargetActivity(self, pharosTargetId):
        try:
            return pharosTargetId in self.__aD
        except Exception:
            return False

    def hasTargetInfo(self, pharosTargetId):
        try:
            return pharosTargetId in self.__tD
        except Exception:
            return False

    def getTargetInfo(self, pharosTargetId, ky):
        try:
            return self.__tD[pharosTargetId][ky]
        except Exception:
            return None

    def fetchTargetActivityData(self):
        targetD = {}
        cofactorFilePath = os.path.join(self.__cachePath, "Pharos-targets", "drug_activity.tdd")
        cfDL = self.__mU.doImport(cofactorFilePath, fmt="tdd", rowFormat="dict")
        targetD = self.__extractCofactorData(cfDL)
        drgIdS = self.__extractDrugIdentifiers(cfDL)
        #
        cofactorFilePath = os.path.join(self.__cachePath, "Pharos-targets", "cmpd_activity.tdd")
        cfDL = self.__mU.doImport(cofactorFilePath, fmt="tdd", rowFormat="dict")
        targetD.update(self.__extractCofactorData(cfDL))
        cmpIdS = self.__extractCompoundIdentifiers(cfDL)
        chemblIdList = list(cmpIdS.union(drgIdS))
        phP = PharosProvider(cachePath=self.__cachePath, useCache=False)
        phP.load(chemblIdList, "identifiers", fmt="json", indent=0)
        #
        targetFilePath = os.path.join(self.__cachePath, "Pharos-targets", "protein.tdd")
        tDL = self.__mU.doImport(targetFilePath, fmt="tdd", rowFormat="dict")
        targetDetailsD = self.__getTargetDetails(tDL)
        #
        pharosReadmePath = os.path.join(self.__cachePath, "Pharos-targets", "pharos-readme.txt")
        readmeLines = self.__mU.doImport(pharosReadmePath, fmt="list")
        self.__version = readmeLines[0].split(" ")[1][1:] if readmeLines else "6"
        #
        tS = datetime.datetime.now().isoformat()
        # vS = datetime.datetime.now().strftime("%Y-%m-%d")
        vS = self.__version
        ok = self.__mU.doExport(self.getTargetActivityDataPath(), {"version": vS, "created": tS, "activity": targetD, "targets": targetDetailsD}, fmt="json", indent=3)
        return ok

    def fetchCompoundIdentifiers(self):
        cofactorFilePath = os.path.join(self.__cachePath, "Pharos-targets", "drug_activity.tdd")
        cfDL = self.__mU.doImport(cofactorFilePath, fmt="tdd", rowFormat="dict")
        drgIdS = self.__extractDrugIdentifiers(cfDL)
        #
        cofactorFilePath = os.path.join(self.__cachePath, "Pharos-targets", "cmpd_activity.tdd")
        cfDL = self.__mU.doImport(cofactorFilePath, fmt="tdd", rowFormat="dict")
        cmpIdS = self.__extractCompoundIdentifiers(cfDL)
        chemblIdList = list(cmpIdS.union(drgIdS))
        return chemblIdList

    def __extractCompoundIdentifiers(self, cfDL):
        chemblIdS = set()
        for cfD in cfDL:
            if cfD["catype"].upper() == "CHEMBL":
                chemblIdS.add(cfD["cmpd_id_in_src"])
        return chemblIdS

    def __extractDrugIdentifiers(self, cfDL):
        chemblIdS = set()
        for cfD in cfDL:
            if cfD["cmpd_chemblid"] and cfD["cmpd_chemblid"].startswith("CHEM"):
                chemblIdS.add(cfD["cmpd_chemblid"])
        return chemblIdS

    def __extractCofactorData(self, cfDL):
        """Extract ids, activity and moa data for drugs and cofactors from the Pharos schema dump files.

        Args:
            cfDL (list): list of dictionaries of containing pharos exported db data.

        Returns:
            dict: dictionary of extracted cofactor data
        """
        try:
            qD = {}
            targetD = {}
            dupD = {}
            for cfD in cfDL:
                tId = cfD["target_id"]
                qD = {}
                qD["smiles"] = cfD["smiles"] if "smiles" in cfD and cfD["smiles"] not in ["N", "NULL"] else None
                qD["chemblId"] = cfD["cmpd_chemblid"] if "cmpd_chemblid" in cfD else None
                qD["chemblId"] = cfD["cmpd_id_in_src"] if "catype" in cfD and cfD["catype"].upper() == "CHEMBL" else qD["chemblId"]
                qD["chemblId"] = qD["chemblId"] if qD["chemblId"] not in ["N", "NULL"] else None
                qD["pubChemId"] = cfD["cmpd_pubchem_cid"] if "cmpd_pubchem_cid" in cfD and cfD["cmpd_pubchem_cid"] not in ["NULL"] else None
                #
                qD["activity"] = cfD["act_value"] if "act_value" in cfD and cfD["act_value"] != "NULL" else None
                qD["activityType"] = cfD["act_type"] if "act_type" in cfD and cfD["act_type"] != "NULL" else None
                if qD["activity"] is not None:
                    qD["activity"] = float(qD["activity"])
                    # qD["activityUnits"] = "nM"
                #
                qD["action"] = cfD["action"] if "action" in cfD and cfD["moa"] == "1" else None
                qD["pharmacology"] = cfD["nlm_drug_info"] if "nlm_drug_info" in cfD else None
                tS = cfD["drug"] if "drug" in cfD else None
                tS = cfD["cmpd_name_in_src"] if "cmpd_name_in_src" in cfD and cfD["cmpd_name_in_src"] != "NULL" else tS
                #
                if tS and tS.startswith("US"):
                    tSL = tS.split(",")
                    qD["patents"] = [t for t in tSL if t.startswith("US")]
                elif tS:
                    # handle the double colon :: separators
                    tSL = tS.split("::")
                    if not tSL:
                        qD["molecule_name"] = tS
                    else:
                        for tS in tSL:
                            if tS.startswith("CHEMBL"):
                                continue
                            elif tS.startswith("US"):
                                ttL = tS.split(",")
                                qD.setdefault("patents", []).append(ttL[0])
                            else:
                                qD["molecule_name"] = tS

                #
                pmId = None
                tS = cfD["reference"] if "reference" in cfD else None
                if tS and "pubmed" in tS:
                    pmId = tS.split("/")[-1]
                tS = cfD["pubmed_ids"].split(",")[0] if "pubmed_ids" in cfD and cfD["pubmed_ids"] else pmId
                qD["pubmedId"] = tS if tS and tS not in ["NULL"] else None
                #
                if qD["activity"] and qD["chemblId"] and (qD["chemblId"], qD["activityType"], qD["action"]) not in dupD:
                    dupD[(qD["chemblId"], qD["activityType"], qD["action"])] = True
                    targetD.setdefault(tId, []).append({ky: qD[ky] for ky in qD if qD[ky] is not None})

            #
        except Exception as e:
            logger.exception("Failing with %r %s", qD, str(e))
        return targetD

    def __getTargetDetails(self, targetDL):
        # Pharos protein target - protein.tdd
        # id	name	description	uniprot	up_version	geneid	sym	family	chr	seq	dtoid	stringid	dtoclass
        rD = {}
        try:
            for targetD in targetDL:
                proteinId = targetD["id"]
                unpId = targetD["uniprot"] if "uniprot" in targetD and targetD["uniprot"] != "NULL" else None
                descr = targetD["description"] if "description" in targetD and targetD["description"] != "NULL" else None
                geneId = targetD["geneid"] if "geneid" in targetD and targetD["geneid"] != "NULL" else None
                dtoId = targetD["dtoid"] if "dtoid" in targetD and targetD["dtoid"] != "NULL" else None
                dtoClass = targetD["dtoclass"] if "dtoclass" in targetD and targetD["dtoclass"] != "NULL" else None
                rD[proteinId] = {"unpId": unpId, "name": descr, "geneId": geneId, "dtoId": dtoId, "dtoClass": dtoClass}
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return rD



##
#  File:           PharosTargetCofactorProvider.py
#  Date:           15-Jun-2021 jdw
#
#  Updated:
#   3-Mar-2023 aae Handle missing activityType in buildCofactorList
#  20-Aug-2024 dwp Add support for loading and accessing data on MongoDB
#
##
"""
Accessors for Pharos target cofactor data.
"""

import datetime
import logging
import os.path
import time

from rcsb.utils.io.FileUtil import FileUtil
from rcsb.utils.io.MarshalUtil import MarshalUtil
from rcsb.utils.io.StashableBase import StashableBase
from rcsb.utils.targets.PharosTargetActivityProvider import PharosTargetActivityProvider
from rcsb.utils.targets.TargetCofactorDbProvider import TargetCofactorDbProvider

logger = logging.getLogger(__name__)


class PharosTargetCofactorProvider(StashableBase):
    """Accessors for Pharos target cofactors."""

    def __init__(self, **kwargs):
        #
        self.__cachePath = kwargs.get("cachePath", ".")
        self.__dirName = "Pharos-cofactors"
        self.__cofactorResourceName = "pharos"
        super(PharosTargetCofactorProvider, self).__init__(self.__cachePath, [self.__dirName])
        self.__dirPath = os.path.join(self.__cachePath, self.__dirName)
        #
        self.__mU = MarshalUtil(workPath=self.__dirPath)
        self.__fD = self.__reload(self.__dirPath, **kwargs)
        #

    def testCache(self, minCount=1):
        logger.info("Pharos cached cofactor count %d", len(self.__fD["cofactors"]) if "cofactors" in self.__fD else 0)
        if self.__fD and "cofactors" in self.__fD and len(self.__fD["cofactors"]) > minCount:
            return True
        else:
            return False

    def hasTarget(self, rcsbEntityId):
        return rcsbEntityId.upper() in self.__fD["cofactors"]

    def getTargets(self, rcsbEntityId):
        try:
            return self.__fD["cofactors"][rcsbEntityId.upper()]
        except Exception:
            return []

    def __getCofactorDataPath(self):
        return os.path.join(self.__dirPath, "Pharos-cofactor-data.json")

    def reload(self):
        self.__fD = self.__reload(self.__dirPath, useCache=True)
        return True

    def __reload(self, dirPath, **kwargs):
        startTime = time.time()
        fD = {}
        useCache = kwargs.get("useCache", True)
        ok = False
        cofactorPath = self.__getCofactorDataPath()
        #
        logger.info("useCache %r cofactorPath %r", useCache, cofactorPath)
        if useCache and self.__mU.exists(cofactorPath):
            fD = self.__mU.doImport(cofactorPath, fmt="json")
            ok = True
        else:
            fU = FileUtil()
            fU.mkdir(dirPath)
        # ---
        numCofactors = len(fD["cofactors"]) if fD and "cofactors" in fD else 0
        logger.info(
            "Completed reload of (%d) cofactors with status (%r) at %s (%.4f seconds)", numCofactors, ok, time.strftime("%Y %m %d %H:%M:%S", time.localtime()), time.time() - startTime
        )
        return fD

    def getCofactorDataDict(self):
        return self.__fD["cofactors"]

    def buildCofactorList(self, sequenceMatchFilePath, crmpObj=None, lnmpObj=None, maxActivity=5):
        """Build target cofactor list for the matching entities in the input sequence match file.

        Args:
            sequenceMatchFilePath (str): sequence match output file path
            crmpObj (obj, optional): instance of ChemRefMappingProviderObj(). Defaults to None.
            lnmpObj (obj, optional): instance of LigandNeighborMappingProviderObj(). Defaults to None.
            maxActivity (int, optional): maximum number of prioritized activity records per target. Defaults to 5.

        Returns:
            bool: True for success or False otherwise

            Example Pharos activity record -

            {
            "version": "2021-06-17",
            "created": "2021-06-17T11:10:54.563394",
            "activity": {
                "2232": [
                    {
                        "smiles": "CC(=CCC\\\\C(=C/Cc1c(O)cc(O)c(C(=O)CCc2ccc(O)cc2)c1O)\\\\C)C",
                        "chemblId": "CHEMBL3360923",
                        "pubChemId": "118724585",
                        "activity": 6.0,
                        "activityType": "IC50",
                        "activityUnits": "nM",
                        "name": "1-[3-(3,7-dimethylocta-2,6-dien-1-yl)-2,4,6-trihydroxyphenyl]-3-(4-hydroxyphenyl)propan-1-one",
                        "pubmedId": "25375026",
                        "patent": "USxxxxxx",
                    }, ...
        """
        rDL = []
        mD = self.__mU.doImport(sequenceMatchFilePath, fmt="json")
        # ---
        chaP = PharosTargetActivityProvider(cachePath=self.__cachePath, useCache=True)
        if not chaP.testCache():
            okP = chaP.fetchTargetActivityData()
            chaP.reload()
            if not chaP.testCache():
                logger.warning("Skipping build of target cofactor list because Pharos Target Activity data is missing (rebuild status %r).", okP)
                return False
        #
        provenanceSource = "Pharos"
        refScheme = "PDB entity"
        assignVersion = chaP.getAssignmentVersion()
        for queryId, matchDL in mD.items():
            # "O43508|uniprotId|7987|proteinId|9606|taxId"
            qCmtD = self.__decodeComment(queryId)
            unpId = qCmtD["uniprotId"]
            queryTaxId = qCmtD["taxId"] if "taxId" in qCmtD else None
            pharosId = qCmtD["proteinId"]
            if queryTaxId == "-1":
                logger.debug("Skipping target with missing taxonomy %r (%r)", unpId, pharosId)
                continue
            #
            if not chaP.hasTargetActivity(pharosId):
                logger.debug("Skipping target with no activities %r (%r)", unpId, pharosId)
                # continue
            # --
            chemCompNeighborsD = {}
            if lnmpObj:
                for matchD in matchDL:
                    tCmtD = self.__decodeComment(matchD["target"])
                    entryId = tCmtD["entityId"].split("_")[0]
                    entityId = tCmtD["entityId"].split("_")[1]
                    rcsbEntityId = entryId + "_" + entityId
                    chemCompIdList = lnmpObj.getLigandNeighbors(rcsbEntityId)
                    chemCompNeighborsD.update({k: True for k in chemCompIdList})
            # --
            queryName = chaP.getTargetInfo(pharosId, "name")
            # --
            for matchD in matchDL:
                tCmtD = self.__decodeComment(matchD["target"])
                entryId = tCmtD["entityId"].split("_")[0]
                entityId = tCmtD["entityId"].split("_")[1]
                rcsbEntityId = entryId + "_" + entityId
                #
                taDL = chaP.getTargetActivity(pharosId)
                logger.debug("Target %r has (%d) activity records", pharosId, len(taDL))
                actL = []
                # cfDL = []
                chD = {}
                for taD in taDL:
                    if taD["chemblId"] in chD:
                        chD[taD["chemblId"]] = True
                        continue

                    actD = {
                        "cofactor_id": taD["chemblId"],
                        "cofactor_name": taD["molecule_name"] if "name" in taD else None,
                        "measurement_type": "p" + taD["activityType"] if "activityType" in taD else None,
                        "measurement_value": taD["activity"],
                        "pubmed_ids": [taD["pubmedId"]] if "pubmedId" in taD else None,
                        "patent_nos": taD["patents"] if "patents" in taD else None,
                        "smiles": taD["smiles"] if "smiles" in taD else None,
                        "action": taD["action"] if "action" in taD else None,
                        "pharmacology": taD["pharmacology"] if "pharmacology" in taD else None,
                    }
                    actD = self.__addLocalIds(actD, crmpObj=crmpObj)
                    actL.append(actD)
                #
                actL = self.__activityListSelect(actL, chemCompNeighborsD, maxActivity=maxActivity)
                if not actL:
                    logger.debug("No Pharos cofactors for %s %s", pharosId, unpId)
                # ---
                # aligned_target.entity_beg_seq_id (current target is PDB entity in json)
                # aligned_target.target_beg_seq_id (current query is target seq in json)
                # aligned_target.length
                fpL = []
                if "alignedRegions" in matchD:
                    fpL = [
                        {
                            "entity_beg_seq_id": arD["targetBegin"],
                            "target_beg_seq_id": arD["queryBegin"],
                            "length": arD["targetEnd"] - arD["targetBegin"],
                        }
                        for arD in matchD["alignedRegions"]
                    ]
                else:
                    fpL = [
                        {
                            "entity_beg_seq_id": matchD["targetBegin"],
                            "target_beg_seq_id": matchD["queryBegin"],
                            "length": matchD["alignLen"],
                        }
                    ]
                # ---
                rD = {
                    "entry_id": entryId,
                    "entity_id": entityId,
                    "query_uniprot_id": unpId,
                    "query_id": pharosId,
                    "query_id_type": "Pharos",
                    "query_name": queryName,
                    "provenance_source": provenanceSource,
                    "reference_scheme": refScheme,
                    "assignment_version": assignVersion,
                    "query_taxonomy_id": int(queryTaxId) if queryTaxId else None,
                    "target_taxonomy_id": int(matchD["targetTaxId"]) if "targetTaxId" in matchD else None,
                    "aligned_target": fpL,
                    "taxonomy_match_status": matchD["taxonomyMatchStatus"] if "taxonomyMatchStatus" in matchD else None,
                    "lca_taxonomy_id": matchD["lcaTaxId"] if "lcaTaxId" in matchD else None,
                    "lca_taxonomy_name": matchD["lcaTaxName"] if "lcaTaxName" in matchD else None,
                    "lca_taxonomy_rank": matchD["lcaRank"] if "lcaRank" in matchD else None,
                    "cofactors": actL,
                }
                rDL.append(rD)
        #
        qD = {}
        for rD in rDL:
            eId = rD["entry_id"] + "_" + rD["entity_id"]
            qD.setdefault(eId, []).append(rD)
        #
        fp = self.__getCofactorDataPath()
        tS = datetime.datetime.now().isoformat()
        # vS = datetime.datetime.now().strftime("%Y-%m-%d")
        vS = assignVersion
        ok = self.__mU.doExport(fp, {"version": vS, "created": tS, "cofactors": qD}, fmt="json", indent=3)
        return ok

    def __addLocalIds(self, cfD, crmpObj=None):
        #
        if crmpObj:
            localIdL = crmpObj.getLocalIds("CHEMBL", cfD["cofactor_id"])
            if localIdL:
                localId = localIdL[0]
                if localId.startswith("PRD_"):
                    cfD["prd_id"] = localId
                else:
                    cfD["chem_comp_id"] = localId
        return cfD

    def __activityListSelect(self, activityDL, chemCompNeighborsD, maxActivity=5):
        """Prioritizing the activity data for locally mapped neighbor ligands and the best binding examples.

        Args:
            activityDL (list): full list of activity objects
            chemCompNeighborsD (dict, optional): index of all chemical components with neighbor interactions to the query target. Defaults {}.
            maxCount (int, optional): maximum number of activity object returned. Defaults to 5.

        Returns:
            list: prioritized and trimmed list of activity objects
        """
        retL = []
        mappedNeighborL = []
        unmappedL = activityDL

        if chemCompNeighborsD:
            unmappedL = []
            # Select out the any cases for molecules that map to a neighbor chemical component.
            for activityD in activityDL:
                if "chem_comp_id" in activityD and activityD["chem_comp_id"] in chemCompNeighborsD:
                    activityD["neighbor_in_pdb"] = "Y"
                    mappedNeighborL.append(activityD)
                else:
                    unmappedL.append(activityD)
                    activityD["neighbor_in_pdb"] = "N"
        #
        numLeft = maxActivity - len(mappedNeighborL)
        if numLeft > 0:
            unmappedL = sorted(unmappedL, key=lambda k: k["measurement_value"], reverse=True)
            retL = mappedNeighborL
            retL.extend(unmappedL[:numLeft])
            retL = sorted(retL, key=lambda k: k["measurement_value"], reverse=True)
        else:
            logger.debug("Mapped neighbor cofactors (%d) excluded unmapped (%d)", len(mappedNeighborL), len(unmappedL))
            retL = sorted(mappedNeighborL, key=lambda k: k["measurement_value"], reverse=True)

        return retL

    def __decodeComment(self, comment, separator="|"):
        dD = {}
        try:
            ti = iter(comment.split(separator))
            dD = {tup[1]: tup[0] for tup in zip(ti, ti)}
        except Exception:
            pass
        return dD

    def loadCofactorData(self, cfgOb, **kwargs):
        """Load cofactor data to MongoDB.
        """
        ok = False
        try:
            tcDbP = TargetCofactorDbProvider(
                cachePath=self.__cachePath,
                cfgOb=cfgOb,
                cofactorResourceName=self.__cofactorResourceName,
                **kwargs
            )
            ok = tcDbP.loadCofactorData(self.__cofactorResourceName, self)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        #
        logger.info("%r cofactor data DB load status (%r)", self.__cofactorResourceName, ok)
        return ok


class PharosTargetCofactorAccessor:
    def __init__(self, cachePath, cfgOb=None, **kwargs):
        """
        Accessor class for fetching cofactor data from MongoDB.
        """
        self.__cofactorResourceName = "pharos"
        self.__cfgOb = cfgOb
        self.__cachePath = cachePath
        #
        self.__tcDbP = TargetCofactorDbProvider(
            cachePath=self.__cachePath,
            cfgOb=self.__cfgOb,
            cofactorResourceName=self.__cofactorResourceName,
            **kwargs
        )

    def testCache(self, minCount=0):
        docCount = self.__tcDbP.cofactorDbCount()
        logger.info("Loaded %s cofactor DB count %d", self.__cofactorResourceName, docCount)
        return docCount >= minCount

    def reload(self):
        return True

    def getTargets(self, rcsbEntityId, dataFieldName="rcsb_cofactors", **kwargs):
        return self.__tcDbP.fetchCofactorData(rcsbEntityId, dataFieldName, **kwargs)



##
#  File:           CARDTargetAnnotationProvider.py
#  Date:           6-Mar-2023 dwp
#
#  Updates:
#  27-Aug-2024 dwp Update testCache and usage of CARDTargetOntologyProvider
#   5-Jun-2025 dwp Update buildAnnotationList to use try/except in case of error
#
##
"""
Accessors and generators for CARD target annotation data.
"""

import datetime
import logging
import os.path
import time

from rcsb.utils.io.FileUtil import FileUtil
from rcsb.utils.io.MarshalUtil import MarshalUtil
from rcsb.utils.io.StashableBase import StashableBase
from rcsb.utils.targets.CARDTargetProvider import CARDTargetProvider
from rcsb.utils.targets.CARDTargetOntologyProvider import CARDTargetOntologyProvider

logger = logging.getLogger(__name__)


class CARDTargetAnnotationProvider(StashableBase):
    """Accessors and generators for CARD target annotation data."""

    def __init__(self, **kwargs):
        #
        self.__cachePath = kwargs.get("cachePath", ".")
        self.__dirName = "CARD-annotations"
        super(CARDTargetAnnotationProvider, self).__init__(self.__cachePath, [self.__dirName])
        self.__dirPath = os.path.join(self.__cachePath, self.__dirName)
        #
        self.__mU = MarshalUtil(workPath=self.__dirPath)
        self.__fD = self.__reload(self.__dirPath, **kwargs)
        #

    def testCache(self, minCount=590):
        ok = False
        if self.__fD and "annotations" in self.__fD and len(self.__fD["annotations"]) > minCount:
            ok = True
        logger.info("CARD annotation testCache status (%r) count %d", ok, len(self.__fD["annotations"]) if "annotations" in self.__fD else 0)
        return ok

    def hasAnnotation(self, rcsbEntityId):
        return rcsbEntityId.upper() in self.__fD["annotations"]

    def getAnnotation(self, rcsbEntityId):
        try:
            return self.__fD["annotations"][rcsbEntityId.upper()]
        except Exception:
            pass
        return None

    def reload(self):
        self.__fD = self.__reload(self.__dirPath, useCache=True)
        ok = self.testCache()
        return ok

    def __getAnnotationDataPath(self):
        return os.path.join(self.__dirPath, "CARD-annotation-data.json")

    def __reload(self, dirPath, **kwargs):
        startTime = time.time()
        fD = {}
        useCache = kwargs.get("useCache", True)
        ok = False
        annotationPath = self.__getAnnotationDataPath()
        #
        logger.info("useCache %r annotationPath %r", useCache, annotationPath)
        if useCache and self.__mU.exists(annotationPath):
            fD = self.__mU.doImport(annotationPath, fmt="json")
            ok = True
        else:
            fU = FileUtil()
            fU.mkdir(dirPath)
        # ---
        logger.info("Completed reload (%r) at %s (%.4f seconds)", ok, time.strftime("%Y %m %d %H:%M:%S", time.localtime()), time.time() - startTime)
        return fD

    def buildAnnotationList(self, sequenceMatchFilePath, useTaxonomy=False):
        """Build polymer entity annotation list for the matching entities in the input sequence match file.

        Args:
            sequenceMatchFilePath (str): sequence match output file path
            useTaxonomy (bool): apply taxonomy criteria to filter matches (only include matches that have a defined taxonomyMatchStatus). Defaults to False

        Returns:
            bool: True for success or False otherwise
        """
        rDL = []
        cardP = CARDTargetProvider(cachePath=self.__cachePath, useCache=False)
        if not cardP.testCache():
            logger.warning("Skipping build of polymer entity annotation list because CARD Target data is missing.")
            return False
        ontologyP = CARDTargetOntologyProvider(cachePath=self.__cachePath, useCache=True)
        if not ontologyP.testCache():
            ok = ontologyP.buildOntologyData()
            ontologyP.reload()
            if not (ok and ontologyP.testCache()):
                logger.warning("Skipping build of polymer entity annotation list because CARD Target Ontology data is missing.")
                return False
        mD = self.__mU.doImport(sequenceMatchFilePath, fmt="json")
        #
        refScheme = "PDB entity"
        assignVersion = cardP.getAssignmentVersion()
        for queryId, matchDL in mD.items():
            qCmtD = self.__decodeComment(queryId)
            modelId = qCmtD["modelId"]
            if not cardP.hasModel(modelId):
                logger.info("Skipping CARD model %r", modelId)
                continue

            for matchD in matchDL:
                try:
                    #
                    tCmtD = self.__decodeComment(matchD["target"])
                    entryId = tCmtD["entityId"].split("_")[0]
                    entityId = tCmtD["entityId"].split("_")[1]
                    seqIdPct = matchD["sequenceIdentity"]
                    bitScore = matchD["bitScore"]
                    nm = cardP.getModelValue(modelId, "modelName")
                    descr = cardP.getModelValue(modelId, "descr")
                    cvTermId = cardP.getModelValue(modelId, "cvTermId")
                    annotationId = "ARO:" + cardP.getModelValue(modelId, "accession")
                    annotationLineage = ontologyP.getLineage(annotationId)
                    familyCvTermId = cardP.getModelValue(modelId, "familyCvTermId")
                    familyName = cardP.getModelValue(modelId, "familyName")
                    familyAnnotationId = "ARO:" + cardP.getModelValue(modelId, "familyAccession")
                    familyDescription = cardP.getModelValue(modelId, "familyDescription")
                    drugClasses = cardP.getModelValue(modelId, "drugClasses")
                    resistanceMechanism = cardP.getModelValue(modelId, "resistanceMechanism")
                    familyAnnotationLineage = ontologyP.getLineage(familyAnnotationId)
                    rD = {
                        "entry_id": entryId,
                        "entity_id": entityId,
                        # "type": "CARD",  # Assigned in DictMethodEntityHelper
                        # "provenance_source": "matching CARD Protein Homolog Models (PHM)",  # Assigned in DictMethodEntityHelper
                        "name": nm,
                        "annotation_id": annotationId,
                        "card_aro_cvterm_id": cvTermId,
                        "description": descr,
                        "seq_id_pct": seqIdPct,
                        "bit_score": bitScore,
                        "reference_scheme": refScheme,
                        "assignment_version": assignVersion,
                        "annotation_lineage": annotationLineage,
                        "query_tax_name": matchD["queryTaxName"] if "queryTaxName" in matchD else None,
                        "target_tax_name": matchD["targetTaxName"] if "targetTaxName" in matchD else None,
                        "match_status": matchD["taxonomyMatchStatus"] if "taxonomyMatchStatus" in matchD else None,
                        "lca_tax_name": matchD["lcaTaxName"] if "lcaTaxName" in matchD else None,
                        "family_annotation_id": familyAnnotationId,
                        "family_card_aro_cvterm_id": familyCvTermId,
                        "family_name": familyName,
                        "family_description": familyDescription,
                        "card_aro_drug_class": drugClasses,
                        "card_aro_resistance_mechanism": resistanceMechanism,
                        "family_annotation_lineage": familyAnnotationLineage,
                    }
                    rDL.append(rD)
                except Exception as e:
                    logger.warning("Unable to process match data for modelId %r with exception %r", modelId, e)
        #
        qD = {}
        dD = {}
        for rD in rDL:
            eId = rD["entry_id"] + "_" + rD["entity_id"]
            fId = rD["annotation_id"]
            if (eId, fId) in dD:
                continue
            dD[(eId, fId)] = True
            qD.setdefault(eId, []).append(rD)
        # --
        # If useTaxonomy filter is True, resulting dictionary contains ~50 fewer perfect matches (as of March 2023)
        if useTaxonomy:
            fqD = {}
            for eId, rDL in qD.items():
                mL = []
                oL = []
                for rD in rDL:
                    if "match_status" not in rD:
                        continue
                    if rD["match_status"] in ["matched", "alternate strain", "alternate strain/serotype/isolate/genotype", "query is ancestor", "query is descendant"]:
                        mL.append(rD)
                    elif rD["match_status"] == "orthologous match (by lca)":
                        oL.append(rD)
                    else:
                        continue
                if not mL and not oL:
                    continue
                logger.info("eId %r mL (%d) oL (%d)", eId, len(mL), len(oL))
                fqD.setdefault(eId, []).extend(mL if mL else oL)
        else:
            fqD = qD
        # --
        # Determine whether entities demonstrate a perfect match or match multiple CARD annotations
        cqD = {}
        for eId, rDL in fqD.items():
            if len(rDL) == 1:
                rD = rDL[0]
                rD["perfect_match"] = "Y"
                cqD[eId] = rD
            else:
                # Sort list of matches by highest sequence identity percent and bit score, as well as annotation_id as a tie breaker
                rDLSorted = sorted(rDL, key=lambda k: (-k["seq_id_pct"], -k["bit_score"], k["annotation_id"]))
                rD = rDLSorted[0]
                # Check for rD that has 100% match, and if so, set this a perfect_match
                if rD["seq_id_pct"] == 100.0:
                    rD["perfect_match"] = "Y"
                else:
                    rD["perfect_match"] = "N"
                cqD[eId] = rD
        # --
        fp = self.__getAnnotationDataPath()
        tS = datetime.datetime.now().isoformat()
        vS = datetime.datetime.now().strftime("%Y-%m-%d")
        ok = self.__mU.doExport(fp, {"version": vS, "created": tS, "taxonomyFilter": useTaxonomy, "annotations": cqD}, fmt="json", indent=3)
        logger.info("Done rebuilding CARD annotation features (status %r) version (%r) created (%r) useTaxonomy (%r) annotations (%d)", ok, vS, tS, useTaxonomy, len(cqD))
        return ok

    def __decodeComment(self, comment, separator="|"):
        dD = {}
        try:
            ti = iter(comment.split(separator))
            dD = {tup[1]: tup[0] for tup in zip(ti, ti)}
        except Exception:
            pass
        return dD



##
# File: DrugBankProvider.py
# Author:  J. Westbrook
# Date:  8-Nov-2020
#
# Manage access to DrugBank target data.
#
# Update:
#  22-Jun-2021 jdw replaced the download of FASTA files with processing the main DrugBank download data file.
#                  Leaving the deprecated FASTA code for now -
#
##

__docformat__ = "google en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import datetime
import logging
import os
import time

from rcsb.utils.chemref.DrugBankProvider import DrugBankProvider
from rcsb.utils.io.FileUtil import FileUtil
from rcsb.utils.io.MarshalUtil import MarshalUtil
from rcsb.utils.seq.UniProtIdMappingProvider import UniProtIdMappingProvider

logger = logging.getLogger(__name__)


class DrugBankTargetProvider(object):
    """Utilities to manage DrugBank target FASTA data.

    Note: UniProt mapping URL template:

        https://go.drugbank.com/polypeptides/P23458
    """

    def __init__(self, **kwargs):
        #
        self.__cachePath = kwargs.get("cachePath", ".")
        self.__dirPath = os.path.join(self.__cachePath, "DrugBank-targets")
        self.__cfD = None
        # self.__fastaPathList = self.__reloadFasta(self.__dirPath, **kwargs)
        self.__dbP = self.__reloadDrugBankProvider(**kwargs)
        self.__version = self.__dbP.getVersion() if self.__dbP else None
        self.__mU = MarshalUtil(workPath=self.__dirPath)

    def testCache(self):
        try:
            # return len(self.__fastaPathList) >= 4
            return self.__dbP.testCache()
        except Exception as e:
            logger.debug("Failing with %s", str(e))
        return False

    def getAssignmentVersion(self):
        return self.__version if self.__version else datetime.datetime.now().strftime("%Y-%m-%d")

    def getFastaPath(self):
        return os.path.join(self.__cachePath, "FASTA", "drugbank")

    def __reloadFasta(self, dirPath, **kwargs):
        """Reload DrugBank target FASTA data files.

        Args:
            dirPath (str, optional): path to DrugBank cache directory
            useCache (bool, optional): flag to use cached files. Defaults to True.

        Returns:

        """
        startTime = time.time()
        logger.info("Starting db reload at %s", time.strftime("%Y %m %d %H:%M:%S", time.localtime()))
        retFilePathList = []
        urlTargetL = [
            "https://go.drugbank.com/releases/latest/downloads/target-all-polypeptide-sequences",
            "https://go.drugbank.com/releases/latest/downloads/enzyme-all-polypeptide-sequences",
            "https://go.drugbank.com/releases/latest/downloads/carrier-all-polypeptide-sequences",
            "https://go.drugbank.com/releases/latest/downloads/transporter-all-polypeptide-sequences",
        ]

        useCache = kwargs.get("useCache", True)
        username = kwargs.get("username", None)
        password = kwargs.get("password", None)
        #
        if not username or not password:
            return retFilePathList
        #
        fU = FileUtil()
        fU.mkdir(dirPath)
        #
        if not useCache:
            #  Clear any cached files
            for urlTarget in urlTargetL:
                baseFileName = fU.getFileName(urlTarget)
                zipFileName = baseFileName + ".fasta.zip"
                retFileName = baseFileName + ".fa"
                for fn in [baseFileName, zipFileName, retFileName]:
                    try:
                        fp = os.path.join(dirPath, fn)
                        os.remove(fp)
                    except Exception:
                        pass
        #
        ok = False
        if useCache:
            ok = True
            for urlTarget in urlTargetL:
                baseFileName = fU.getFileName(urlTarget)
                retFileName = baseFileName + ".fa"
                retFilePath = os.path.join(dirPath, retFileName)
                ok = fU.exists(retFilePath)
                if not ok:
                    break
                retFilePathList.append(retFilePath)
        #
        logger.info("Using cached files %r", ok)
        if not useCache or not ok:
            if not username or not password:
                logger.warning("Missing credentials for DrugBank file download...")

            for urlTarget in urlTargetL:
                baseFileName = fU.getFileName(urlTarget)
                zipFileName = baseFileName + ".fasta.zip"
                retFileName = baseFileName + ".fa"
                zipFilePath = os.path.join(dirPath, zipFileName)
                retFilePath = os.path.join(dirPath, retFileName)
                basePath = os.path.join(dirPath, baseFileName)
                logger.info("Fetching url %s for FASTA target file %s", urlTarget, baseFileName)
                ok = fU.get(urlTarget, zipFilePath, username=username, password=password)
                endTime = time.time()
                logger.info("Completed db fetch at %s (%.4f seconds)", time.strftime("%Y %m %d %H:%M:%S", time.localtime()), endTime - startTime)
                #
                ok = fU.unbundleZipfile(zipFilePath, dirPath=basePath)
                fU.put(os.path.join(basePath, "protein.fasta"), retFilePath)
                endTime = time.time()
                logger.info("Completed unzip at %s (%.4f seconds)", time.strftime("%Y %m %d %H:%M:%S", time.localtime()), endTime - startTime)
                retFilePathList.append(retFilePath)
        return retFilePathList

    def exportFasta(self, fastaPath, taxonPath, addTaxonomy=False):
        # ok = self.__consolidateFasta(fastaPath, taxonPath, self.__fastaPathList, addTaxonomy=addTaxonomy)
        ok = self.__buildResourceFiles(fastaPath, taxonPath, addTaxonomy=addTaxonomy)
        return ok

    def __consolidateFasta(self, fastaPath, taxonPath, inpPathList, addTaxonomy=False):
        #
        drugBankTargetMapPath = self.__getTargetDrugMapPath()
        #
        oD = {}
        uD = {}
        taxonL = []
        try:
            if addTaxonomy:
                umP = UniProtIdMappingProvider(self.__cachePath)
                umP.reload(useCache=True)
            #
            for fp in inpPathList:
                fD = self.__mU.doImport(fp, fmt="fasta", commentStyle="default")
                for seqId, sD in fD.items():
                    tL = seqId[seqId.find("(") + 1 : seqId.find(")")]
                    dbIdL = [v.strip() for v in tL.split(";")]
                    seq = sD["sequence"]
                    tL = seqId.split("|")
                    unpId = tL[1].split(" ")[0]
                    cD = {"sequence": seq, "uniprotId": unpId}
                    if addTaxonomy:
                        taxId = umP.getMappedId(unpId, mapName="NCBI-taxon")
                        cD["taxId"] = taxId if taxId else -1
                    #
                    seqId = ""
                    cL = []
                    for k, v in cD.items():
                        if k in ["sequence"]:
                            continue
                        cL.append(str(v))
                        cL.append(str(k))
                    seqId = "|".join(cL)
                    oD[seqId] = cD
                    if addTaxonomy:
                        taxonL.append("%s\t%s" % (seqId, taxId))
                    uD.setdefault(unpId, []).extend(dbIdL)

            ok1 = self.__mU.doExport(fastaPath, oD, fmt="fasta", makeComment=True)
            tS = datetime.datetime.now().isoformat()
            vS = datetime.datetime.now().strftime("%Y-%m-%d")
            ok2 = self.__mU.doExport(drugBankTargetMapPath, {"version": vS, "created": tS, "cofactors": uD}, fmt="json")
            ok3 = True
            if addTaxonomy:
                ok3 = self.__mU.doExport(taxonPath, taxonL, fmt="list")
            return ok1 & ok2 & ok3
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        #
        return False

    def __getTargetDrugMapPath(self):
        return os.path.join(self.__dirPath, "drugbank_target_drug_map.json")

    def hasCofactor(self, unpId):
        if not self.__cfD:
            self.__cfD = self.__reloadCofactors()
        try:
            return unpId in self.__cfD
        except Exception:
            return False

    def getCofactors(self, unpId):
        if not self.__cfD:
            self.__cfD = self.__reloadCofactors()
        try:
            return self.__cfD[unpId]
        except Exception:
            return []

    def __reloadCofactors(self):
        try:
            qD = self.__mU.doImport(self.__getTargetDrugMapPath(), fmt="json")
            self.__version = qD["version"]
            return qD["cofactors"]
        except Exception as e:
            logger.error("Failing with %s", str(e))
        return {}

    def __reloadDrugBankProvider(self, **kwargs):
        useCache = kwargs.get("useCache", True)
        un = kwargs.get("username", None)
        pw = kwargs.get("password", None)
        return DrugBankProvider(cachePath=self.__cachePath, useCache=useCache, username=un, password=pw)

    def __buildResourceFiles(self, fastaPath, taxonPath, addTaxonomy=False):
        """Build DrugBank FASTA and resource files from the full DrugBank XML data download."""
        #
        try:
            drugBankTargetMapPath = self.__getTargetDrugMapPath()
            self.__version = self.__dbP.getVersion()
            #
            if addTaxonomy:
                umP = UniProtIdMappingProvider(self.__cachePath)
                umP.reload(useCache=True)
            #
            oD = {}
            uD = {}
            taxonD = {}
            #
            dbIdL = self.__dbP.getDrugbankIds()
            logger.info("Building resource file for (%d) DrugBank entries", len(dbIdL))
            for dbId in dbIdL:
                tiDL = self.__dbP.getFeature(dbId, "target_interactions")
                for tiD in tiDL:
                    dD = {}
                    dD["drugbank_id"] = dbId
                    dD["name"] = self.__dbP.getFeature(dbId, "name")
                    dD["description"] = self.__dbP.getFeature(dbId, "description")
                    dD["moa"] = self.__dbP.getFeature(dbId, "mechanism-of-action")
                    dD["pharmacology"] = self.__dbP.getFeature(dbId, "pharmacodynamics")
                    dD["inchi_key"] = self.__dbP.getFeature(dbId, "inchikey")
                    dD["smiles"] = self.__dbP.getFeature(dbId, "smiles")
                    dD["pubmed_ids"] = tiD["articles"]
                    tS = tiD["amino-acid-sequence"]
                    if not tS:
                        logger.debug("Skipping entry target %r", tiD)
                        continue
                    sequence = "".join(tS.split("\n")[1:])
                    unpId = tiD["uniprot_ids"]
                    if "," in unpId or ";" in unpId or isinstance(unpId, list):
                        logger.warning("Bad uniprot id %r", unpId)
                    dD["target_name"] = tiD["name"]
                    cD = {"sequence": sequence, "uniprotId": unpId}
                    if addTaxonomy:
                        taxId = umP.getMappedId(unpId, mapName="NCBI-taxon")
                        cD["taxId"] = taxId if taxId else -1
                    #
                    seqId = ""
                    cL = []
                    for k, v in cD.items():
                        if k in ["sequence"]:
                            continue
                        cL.append(str(v))
                        cL.append(str(k))
                    seqId = "|".join(cL)
                    oD[seqId] = cD
                    if addTaxonomy:
                        taxonD["%s\t%s" % (seqId, taxId)] = True
                    #
                    logger.debug("%r dD %r", dbId, dD)
                    uD.setdefault(unpId, []).append(dD)

            ok1 = self.__mU.doExport(fastaPath, oD, fmt="fasta", makeComment=True)
            tS = datetime.datetime.now().isoformat()
            vS = self.__version
            ok2 = self.__mU.doExport(drugBankTargetMapPath, {"version": vS, "created": tS, "cofactors": uD}, fmt="json", indent=3)
            ok3 = True
            if addTaxonomy:
                ok3 = self.__mU.doExport(taxonPath, list(taxonD.keys()), fmt="list")
            return ok1 & ok2 & ok3
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return False

    def __decodeComment(self, comment, separator="|"):
        dD = {}
        try:
            ti = iter(comment.split(separator))
            dD = {tup[1]: tup[0] for tup in zip(ti, ti)}
        except Exception:
            pass
        return dD



##
#  File:           DrugBankTargetCofactorProvider.py
#  Date:           12-Jun-2021 jdw
#
#  Updated:
#  20-Aug-2024 dwp Add support for loading and accessing data on MongoDB
##
"""
Accessors for DrugBank target cofactors.
"""

import datetime
import logging
import os.path
import time

from rcsb.utils.io.FileUtil import FileUtil
from rcsb.utils.io.MarshalUtil import MarshalUtil
from rcsb.utils.io.StashableBase import StashableBase
from rcsb.utils.targets.DrugBankTargetProvider import DrugBankTargetProvider
from rcsb.utils.targets.TargetCofactorDbProvider import TargetCofactorDbProvider

logger = logging.getLogger(__name__)


class DrugBankTargetCofactorProvider(StashableBase):
    """Accessors for DrugBank target cofactors."""

    def __init__(self, **kwargs):
        #
        self.__cachePath = kwargs.get("cachePath", ".")
        self.__useCache = kwargs.get("useCache", True)
        self.__fmt = kwargs.get("fmt", "pickle")
        self.__dirName = "DrugBank-cofactors"
        self.__cofactorResourceName = "drugbank"
        super(DrugBankTargetCofactorProvider, self).__init__(self.__cachePath, [self.__dirName])
        self.__dirPath = os.path.join(self.__cachePath, self.__dirName)
        #
        self.__mU = MarshalUtil(workPath=self.__dirPath)
        self.__fD = self.__reload(self.__dirPath, self.__useCache, self.__fmt)
        #

    def testCache(self, minCount=590):
        logger.info("DrugBank cofactor count %d", len(self.__fD["cofactors"]) if "cofactors" in self.__fD else 0)
        if self.__fD and "cofactors" in self.__fD and len(self.__fD["cofactors"]) > minCount:
            return True
        else:
            return False

    def hasTarget(self, rcsbEntityId):
        return rcsbEntityId.upper() in self.__fD["cofactors"]

    def getTargets(self, rcsbEntityId):
        try:
            return self.__fD["cofactors"][rcsbEntityId.upper()]
        except Exception:
            return []

    def __getCofactorDataPath(self, fmt="json"):
        fExt = "json" if fmt == "json" else "pic"
        return os.path.join(self.__dirPath, "drugbank-cofactor-data.%s" % fExt)

    def reload(self):
        self.__fD = self.__reload(self.__dirPath, useCache=True, fmt=self.__fmt)
        return True

    def __reload(self, dirPath, useCache, fmt):
        startTime = time.time()
        fD = {}

        ok = False
        cofactorPath = self.__getCofactorDataPath(fmt=fmt)
        #
        logger.info("useCache %r cofactorPath %r", useCache, cofactorPath)
        if useCache and self.__mU.exists(cofactorPath):
            fD = self.__mU.doImport(cofactorPath, fmt=fmt)
        else:
            fU = FileUtil()
            fU.mkdir(dirPath)
        # ---
        logger.info("Completed reload (%r) at %s (%.4f seconds)", ok, time.strftime("%Y %m %d %H:%M:%S", time.localtime()), time.time() - startTime)
        return fD

    def getCofactorDataDict(self):
        return self.__fD["cofactors"]

    def buildCofactorList(self, sequenceMatchFilePath, crmpObj=None, lnmpObj=None):
        """Build target cofactor list for the matching entities in the input sequence match file.

        Args:
            sequenceMatchFilePath (str): sequence match output file path
            crmpObj (obj, optional): instance of ChemRefMappingProviderObj(). Defaults to None
            lnmpObj (obj, optional): instance of LigandNeighborMappingProviderObj(). Defaults to None.

        Returns:
            bool: True for success or False otherwise
        """
        rDL = []
        dbP = DrugBankTargetProvider(cachePath=self.__cachePath, useCache=True)
        if not dbP.testCache():
            logger.warning("Skipping build of target cofactor list because DrugBank Target data is missing.")
            return False
        mD = self.__mU.doImport(sequenceMatchFilePath, fmt="json")
        #
        provenanceSource = "DrugBank"
        refScheme = "PDB entity"
        assignVersion = str(dbP.getAssignmentVersion())
        for queryId, matchDL in mD.items():
            qCmtD = self.__decodeComment(queryId)
            unpId = qCmtD["uniprotId"]
            queryTaxId = qCmtD["taxId"] if "taxId" in qCmtD else None
            if not dbP.hasCofactor(unpId) or queryTaxId == "-1":
                logger.info("Skipping target %r", unpId)
                continue
            #
            # --
            chemCompNeighborsD = {}
            if lnmpObj:
                for matchD in matchDL:
                    tCmtD = self.__decodeComment(matchD["target"])
                    entryId = tCmtD["entityId"].split("_")[0]
                    entityId = tCmtD["entityId"].split("_")[1]
                    rcsbEntityId = entryId + "_" + entityId
                    chemCompIdList = lnmpObj.getLigandNeighbors(rcsbEntityId)
                    chemCompNeighborsD.update({k: True for k in chemCompIdList})
            # --
            #
            for matchD in matchDL:
                tCmtD = self.__decodeComment(matchD["target"])
                entryId = tCmtD["entityId"].split("_")[0]
                entityId = tCmtD["entityId"].split("_")[1]
                # --
                dbDL = dbP.getCofactors(unpId)
                # --
                cfDL = []
                for dbD in dbDL:
                    cfD = {}
                    cfD["cofactor_id"] = dbD["drugbank_id"]
                    cfD["molecule_name"] = dbD["name"]
                    cfD["target_name"] = dbD["target_name"]
                    # cfD["description"] = dbD["description"]
                    cfD["moa"] = dbD["moa"]
                    # cfD["pharmacology"] = dbD["pharmacology"]
                    cfD["inchi_key"] = dbD["inchi_key"]
                    cfD["smiles"] = dbD["smiles"]
                    cfD["pubmed_ids"] = dbD["pubmed_ids"]
                    cfD = self.__addLocalIds(cfD, crmpObj)
                    #
                    if "chem_comp_id" in cfD and cfD["chem_comp_id"] in chemCompNeighborsD:
                        cfD["neighbor_in_pdb"] = "Y"
                    else:
                        cfD["neighbor_in_pdb"] = "N"
                    #
                    cfDL.append(cfD)
                # ---
                queryName = cfDL[0]["target_name"] if cfDL and "target_name" in cfDL[0] else None
                # ---
                # aligned_target.entity_beg_seq_id (current target is PDB entity in json)
                # aligned_target.target_beg_seq_id (current query is target seq in json)
                # aligned_target.length
                fpL = []
                if "alignedRegions" in matchD:
                    fpL = [
                        {
                            "entity_beg_seq_id": arD["targetBegin"],
                            "target_beg_seq_id": arD["queryBegin"],
                            "length": arD["targetEnd"] - arD["targetBegin"],
                        }
                        for arD in matchD["alignedRegions"]
                    ]
                else:
                    fpL = [
                        {
                            "entity_beg_seq_id": matchD["targetBegin"],
                            "target_beg_seq_id": matchD["queryBegin"],
                            "length": matchD["alignLen"],
                        }
                    ]
                # ---
                rD = {
                    "entry_id": entryId,
                    "entity_id": entityId,
                    "query_uniprot_id": unpId,
                    "query_id": unpId,
                    "query_id_type": "DrugBank",
                    "query_name": queryName,
                    "provenance_source": provenanceSource,
                    "reference_scheme": refScheme,
                    "assignment_version": assignVersion,
                    "query_taxonomy_id": int(queryTaxId) if queryTaxId else None,
                    "target_taxonomy_id": int(matchD["targetTaxId"]) if "targetTaxId" in matchD else None,
                    "aligned_target": fpL,
                    "taxonomy_match_status": matchD["taxonomyMatchStatus"] if "taxonomyMatchStatus" in matchD else None,
                    "lca_taxonomy_id": matchD["lcaTaxId"] if "lcaTaxId" in matchD else None,
                    "lca_taxonomy_name": matchD["lcaTaxName"] if "lcaTaxName" in matchD else None,
                    "lca_taxonomy_rank": matchD["lcaRank"] if "lcaRank" in matchD else None,
                    "cofactors": cfDL,
                }
                rDL.append(rD)
        #
        qD = {}
        for rD in rDL:
            eId = rD["entry_id"] + "_" + rD["entity_id"]
            qD.setdefault(eId, []).append(rD)
        fp = self.__getCofactorDataPath(fmt=self.__fmt)
        tS = datetime.datetime.now().isoformat()
        # vS = datetime.datetime.now().strftime("%Y-%m-%d")
        vS = assignVersion
        ok = self.__mU.doExport(fp, {"version": vS, "created": tS, "cofactors": qD}, fmt=self.__fmt, indent=3)
        return ok

    def __addLocalIds(self, cfD, crmpOb=None):
        #
        if crmpOb:
            localIdL = crmpOb.getLocalIds("DRUGBANK", cfD["cofactor_id"])
            if localIdL:
                localId = localIdL[0]
                if localId.startswith("PRD_"):
                    cfD["prd_id"] = localId
                else:
                    cfD["chem_comp_id"] = localId
        return cfD

    def __decodeComment(self, comment, separator="|"):
        dD = {}
        try:
            ti = iter(comment.split(separator))
            dD = {tup[1]: tup[0] for tup in zip(ti, ti)}
        except Exception:
            pass
        return dD

    def loadCofactorData(self, cfgOb, **kwargs):
        """Load cofactor data to MongoDB.
        """
        ok = False
        try:
            tcDbP = TargetCofactorDbProvider(
                cachePath=self.__cachePath,
                cfgOb=cfgOb,
                cofactorResourceName=self.__cofactorResourceName,
                **kwargs
            )
            ok = tcDbP.loadCofactorData(self.__cofactorResourceName, self)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        #
        logger.info("%r cofactor data DB load status (%r)", self.__cofactorResourceName, ok)
        return ok


class DrugBankTargetCofactorAccessor:
    def __init__(self, cachePath, cfgOb=None, **kwargs):
        """
        Accessor class for fetching cofactor data from MongoDB.
        """
        self.__cofactorResourceName = "drugbank"
        self.__cfgOb = cfgOb
        self.__cachePath = cachePath
        #
        self.__tcDbP = TargetCofactorDbProvider(
            cachePath=self.__cachePath,
            cfgOb=self.__cfgOb,
            cofactorResourceName=self.__cofactorResourceName,
            **kwargs
        )

    def testCache(self, minCount=0):
        docCount = self.__tcDbP.cofactorDbCount()
        logger.info("Loaded %s cofactor DB count %d", self.__cofactorResourceName, docCount)
        return docCount >= minCount

    def reload(self):
        return True

    def getTargets(self, rcsbEntityId, dataFieldName="rcsb_cofactors", **kwargs):
        return self.__tcDbP.fetchCofactorData(rcsbEntityId, dataFieldName, **kwargs)



##
#  File:           TargetCofactorDbProvider.py
#  Date:           20-Aug-2024 dwp
#
#  Updated:
#  23-Aug-2024 dwp Use context manager for opening connections to ensure they close properly
#  26-Aug-2024 dwp Revert back to persistent client connections
#
##
"""
Provider/accessors for target cofactors data to/from MongoDB.
"""

import logging
import time

from rcsb.db.mongo.Connection import Connection
from rcsb.db.mongo.DocumentLoader import DocumentLoader

logger = logging.getLogger(__name__)


class TargetCofactorDbProvider:
    def __init__(self, cachePath, cfgOb=None, cofactorResourceName=None, **kwargs):
        """
        Provider class for loading and fetching cofactor data to/from MongoDB.
        """
        self.__cachePath = cachePath
        self.__cfgOb = cfgOb
        self.__cofactorResourceName = cofactorResourceName  # "chembl", "pharos", or "drugbank"
        self.__numProc = kwargs.get("numProc", 6)
        self.__chunkSize = kwargs.get("chunkSize", 10)
        self.__resourceName = "MONGO_DB"
        #
        self.__databaseName = kwargs.get("databaseName", "cofactor_exdb")
        self.__collectionName = kwargs.get("collectionName", self.__cofactorResourceName)
        #
        conn = Connection(cfgOb=self.__cfgOb, resourceName=self.__resourceName)
        conn.openConnection()
        self.__client = conn.getClientConnection()
        self.__db = self.__client[self.__databaseName]
        self.__collection = self.__db[self.__collectionName]

    def cofactorDbCount(self):
        """Count the number of documents in the cofactor collection.
        """
        count = 0
        try:
            count = self.__collection.count_documents({})
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return count

    def fetchCofactorData(self, rcsbEntityId, dataFieldName="rcsb_cofactors"):
        """Fetches the 'rcsb_cofactors' field for a document with the given rcsb_id.

        :param rcsb_id: The rcsb_id of the document to fetch.
        :return: The 'rcsb_cofactors' field if present, otherwise None.
        """
        data = []
        try:
            document = self.__collection.find_one({"rcsb_id": rcsbEntityId.upper()})
            if document and dataFieldName in document:
                data = document[dataFieldName]
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return data

    def loadCofactorData(self, cofactorResourceName, cofactorProvider, **kwargs):
        """Load cofactor data for the input data resource.

        Args:
            cofactorResourceName (str): target cofactor resource name (e.g., "pharos", "chembl", or "drugbank).
            cofactorProvider (obj): target cofactor resource provider instance.

        Returns:
            bool: True for success or False otherwise
        """
        if cofactorResourceName not in ["chembl", "pharos", "drugbank"]:
            logger.error("Unsupported cofactor resource %r", cofactorResourceName)
            return False
        #
        loadType = kwargs.get("loadType", "full")
        maxStepLength = kwargs.get("maxStepLength", 10000)
        documentLimit = kwargs.get("documentLimit", None)
        verbose = kwargs.get("verbose", False)
        readBackCheck = kwargs.get("readBackCheck", False)
        #
        dl = DocumentLoader(
            self.__cfgOb,
            self.__cachePath,
            self.__resourceName,
            numProc=self.__numProc,
            chunkSize=self.__chunkSize,
            maxStepLength=maxStepLength,
            documentLimit=documentLimit,
            verbose=verbose,
            readBackCheck=readBackCheck,
        )
        #
        cofactorDataDict = cofactorProvider.getCofactorDataDict()
        dL = []
        dL = [{"rcsb_id": k.upper(), "rcsb_cofactors": vL} for k, vL in cofactorDataDict.items()]
        #
        startTime = time.time()
        ok = dl.load(self.__databaseName, self.__collectionName, loadType=loadType, documentList=dL, indexAttributeList=["rcsb_id"], keyNames=None, schemaLevel=None)
        logger.info(
            "Completed loading cofactor data for %s (%r) to databaseName %s collectionName %s (status %r) at %s (%.4f seconds)",
            cofactorResourceName,
            len(dL),
            self.__databaseName,
            self.__collectionName,
            ok,
            time.strftime("%Y %m %d %H:%M:%S", time.localtime()),
            time.time() - startTime,
        )
        #
        return ok



##
#  File:           IMGTTargetFeatureProvider.py
#  Date:           7-Jul-2021 jdw
#
#  Updated:
#
##
"""
Accessors for IMGT (The International Immunogenetic Information System) target features
"""

import datetime
import logging
import os.path
import time

from rcsb.utils.io.FileUtil import FileUtil
from rcsb.utils.io.MarshalUtil import MarshalUtil
from rcsb.utils.io.StashableBase import StashableBase
from rcsb.utils.targets.IMGTTargetProvider import IMGTTargetProvider

logger = logging.getLogger(__name__)


class IMGTTargetFeatureProvider(StashableBase):
    """Accessors for IMGT (The International Immunogenetic Information System) target features"""

    # Link out using the IMGT -
    # https://www.imgt.org/3Dstructure-DB/cgi/details.cgi?pdbcode=5w5m&Part=Chain
    #
    def __init__(self, **kwargs):
        #
        self.__cachePath = kwargs.get("cachePath", ".")
        useCache = kwargs.get("useCache", True)
        self.__dirName = "IMGT-features"
        super(IMGTTargetFeatureProvider, self).__init__(self.__cachePath, [self.__dirName])
        self.__dirPath = os.path.join(self.__cachePath, self.__dirName)
        #
        self.__mU = MarshalUtil(workPath=self.__dirPath)
        self.__fD = self.__reload(self.__dirPath, useCache)
        #

    def testCache(self, minCount=20000):
        logger.info("IMGT feature count %d", len(self.__fD["features"]) if "features" in self.__fD else 0)
        if self.__fD and "features" in self.__fD and len(self.__fD["features"]) > minCount:
            return True
        else:
            return False

    def hasFeatures(self, rcsbInstanceId):
        """Return if features exist for the input instance identifier (auth_asym_id)

        Args:
            rcsbInstanceId (str): <pdbId (lower case)>.<auth_asym_id (case sensitive)>

        Returns:
            bool: True for success or False otherwise
        """
        return rcsbInstanceId in self.__fD["features"]

    def getFeatures(self, rcsbInstanceId):
        """Return features for the instance identifier (auth_asym_id)

        Args:
            rcsbInstanceId (str): <pdbId (lower case)>.<auth_asym_id (case sensitive)>

        Returns:
            list: list of feature dictionaries
        """
        try:
            return self.__fD["features"][rcsbInstanceId]
        except Exception:
            return []

    def __getFeatureDataPath(self):
        return os.path.join(self.__dirPath, "IMGT-feature-data.json")

    def reload(self):
        self.__fD = self.__reload(self.__dirPath, True)
        return True

    def __reload(self, dirPath, useCache):
        startTime = time.time()
        fD = {}
        featurePath = self.__getFeatureDataPath()
        #
        logger.info("useCache %r featurePath %r", useCache, featurePath)
        if useCache and self.__mU.exists(featurePath):
            fD = self.__mU.doImport(featurePath, fmt="json")
        else:
            fU = FileUtil()
            fU.mkdir(dirPath)
        # ---
        logger.info("Completed reload (useCache %r) at %s (%.4f seconds)", useCache, time.strftime("%Y %m %d %H:%M:%S", time.localtime()), time.time() - startTime)
        return fD

    def buildFeatureList(self, useCache=True):
        """Build polymer instance feature list for IMGT annotations.

        Returns:
            bool: True for success or False otherwise

                    5w5m_B": {
                    "description": "FUSION-TNFRSF1B-GAMMA-1",
                    "domains": {
                        "C-DOMAIN|CH2|1": {
                        "geneAlleles": [
                            {
                                "taxName": "Homo sapiens",
                                "geneAllele": "IGHG4*01"
                            },
                            {
                                "taxName": "Homo sapiens",
                                "geneAllele": "IGHG4*03"
                            },
                            {
                                "taxName": "Homo sapiens",
                                "geneAllele": "IGHG4*04"
                            }
                        ],
                        "alignment": {
                            "begEntitySeqId": 7,
                            "endEntitySeqId": 116,
                            "begIMGTSeqId": "1",
                            "endIMGTSeqId": "105"
                        }
                        },
                        "C-DOMAIN|CH3|2": {
                        "geneAlleles": [
                            {
                                "taxName": "Homo sapiens",
                                "geneAllele": "IGHG4*01"
                            },
                            {
                                "taxName": "Homo sapiens",
                                "geneAllele": "IGHG4*04"
                            }
                        ],
                        "alignment": {
                            "begEntitySeqId": 117,
                            "endEntitySeqId": 221,
                            "begIMGTSeqId": "106",
                            "endIMGTSeqId": "209"
                        }
                        }
                    },
                    "proteinName": "IgG4 Sigma1 Fc",
                    "receptorType": "IG",
                    "receptorDescription": "FUSION-[TNFRSF1B]2-FC-GAMMA-1",
                    "species": "Homo sapiens (human)"
                },
        """
        rDL = []
        imgtP = IMGTTargetProvider(cachePath=self.__cachePath, useCache=useCache)
        if not imgtP.testCache():
            logger.warning("Skipping build of polymer instance feature list because IMGT Target data is missing.")
            return False
        #
        provenanceSource = "IMGT"
        refScheme = "PDB entity"
        assignVersion = imgtP.getVersion()
        #
        chainD = imgtP.getChains()
        #
        fTupL = [
            ("description", "IMGT_ANTIBODY_DESCRIPTION"),
            ("proteinName", "IMGT_ANTIBODY_PROTEIN_NAME"),
            ("receptorType", "IMGT_ANTIBODY_RECEPTOR_TYPE"),
            ("receptorDescription", "IMGT_ANTIBODY_RECEPTOR_DESCRIPTION"),
            ("species", "IMGT_ANTIBODY_ORGANISM_NAME"),
        ]
        ii = 1
        #
        for chainId, chD in chainD.items():
            entryId = chainId[:4]
            authAsymId = chainId.split("_")[1]
            # descriptive features -
            for fTup in fTupL:
                rD = {
                    "entry_id": entryId,
                    "auth_asym_id": authAsymId,
                    "type": fTup[1],
                    "feature_id": "IMGT_" + str(ii),
                    "name": chD[fTup[0]] if fTup[0] in chD else None,
                    "provenance_source": provenanceSource,
                    "reference_scheme": refScheme,
                    "assignment_version": assignVersion,
                    "feature_positions": [],
                }
                rDL.append(rD)
                ii += 1
            # domain features -
            if "domains" not in chD:
                continue
            for domainId, dD in chD["domains"].items():
                dIdL = domainId.split("|")
                domainName = dIdL[0] + " " + dIdL[1]
                begSeqId = endSeqId = None
                if "alignment" in dD:
                    begSeqId = dD["alignment"]["begEntitySeqId"]
                    endSeqId = dD["alignment"]["endEntitySeqId"]
                else:
                    logger.debug("%r missing alignment for in %r", chainId, dD)
                #
                gaL = []
                if "geneAlleles" in dD:
                    for gD in dD["geneAlleles"]:
                        gaL.append(gD["geneAllele"])
                else:
                    logger.debug("%r missing gene and alleles for in %r", chainId, dD)
                #
                #
                rD = {
                    "entry_id": entryId,
                    "auth_asym_id": authAsymId,
                    "type": "IMGT_ANTIBODY_DOMAIN_NAME",
                    "feature_id": "IMGT_" + str(ii),
                    "name": domainName,
                    "provenance_source": provenanceSource,
                    "reference_scheme": refScheme,
                    "assignment_version": assignVersion,
                    "feature_positions": [{"beg_seq_id": begSeqId, "end_seq_id": endSeqId}],
                }
                rDL.append(rD)
                ii += 1
                #
                for ga in gaL:
                    rD = {
                        "entry_id": entryId,
                        "auth_asym_id": authAsymId,
                        "type": "IMGT_ANTIBODY_GENE_ALLELE_NAME",
                        "feature_id": "IMGT_" + str(ii),
                        "name": ga,
                        "provenance_source": provenanceSource,
                        "reference_scheme": refScheme,
                        "assignment_version": assignVersion,
                        "feature_positions": [{"beg_seq_id": begSeqId, "end_seq_id": endSeqId}],
                    }
                    rDL.append(rD)
                    ii += 1
        #
        qD = {}
        for rD in rDL:
            eId = rD["entry_id"] + "." + rD["auth_asym_id"]
            qD.setdefault(eId, []).append(rD)
        #
        logger.info("IMGT antibody chain features (%d)", len(qD))
        #
        fp = self.__getFeatureDataPath()
        tS = datetime.datetime.now().isoformat()
        vS = assignVersion
        ok = self.__mU.doExport(fp, {"version": vS, "created": tS, "features": qD}, fmt="json", indent=3)
        return ok



##
#  File:           SAbDabTargetProvider.py
#  Date:           18-Jun-2021 jdw
#
#  Updated:
#   9-Feb-2023 aae  Find Highest_Clin_Trial column regardless of month
#   1-Jul-2024 dwp  Update SAbDab data parsing (following change in source data headers)
#  22-Jul-2024 dwp  Re-update SAbDab data parsing (following change in source data headers)
##
"""
Accessors for Thera-SAbDab(Therapeutic Structural Antibody Database) target data.
"""

import datetime
import logging
import os.path
import time

from rcsb.utils.io.FileUtil import FileUtil
from rcsb.utils.io.MarshalUtil import MarshalUtil

logger = logging.getLogger(__name__)


class SAbDabTargetProvider(object):
    """Accessors for SAbDab and Thera-SAbDab(Therapeutic Structural Antibody Database) target data.

    See: Dunbar, J., Krawczyk, K. et al (2014). Nucleic Acids Res. 42. D1140-D1146
    """

    def __init__(self, **kwargs):
        #
        self.__cachePath = kwargs.get("cachePath", ".")
        self.__dirPath = os.path.join(self.__cachePath, "SAbDab-features")
        #
        self.__assignVersion = None
        self.__mU = MarshalUtil(workPath=self.__dirPath)
        self.__oD, self.__aD, self.__dumpPath, self.__assignVersion = self.__reload(self.__dirPath, **kwargs)
        #

    def testCache(self, minCount=590):
        logger.info("Therapeutic SAbDab count %d", len(self.__oD))
        if self.__oD and len(self.__oD) > minCount:
            return True
        else:
            return False

    def getFeatures(self, therapeuticName, featureKey):
        fL = []
        try:
            fS = self.__oD[therapeuticName][featureKey]
            if ";" in fS:
                fL = fS.split(";")
            else:
                fL = [fS]
        except Exception:
            fL = []
        return fL

    def getAssignment(self, instanceId, featureKey):
        """Return the value of the key feature for the input instance identifier.

        Args:
            instanceId (str): instance identifier '<pdbId>.<authAsymId>'
            featureKey (str): assignment feature key: one of pdb|Hchain|Lchain|model|antigen_chain|antigen_type|
                              antigen_het_name|antigen_name|heavy_subclass|light_subclass|light_ctype)

        Returns:
            str:  feature value or None
        """
        fVal = None
        try:
            fVal = self.__aD[instanceId][featureKey]
        except Exception:
            fVal = None
        return fVal

    def hasAssignment(self, instanceId):
        """Return if assignment data is available for the input instance.

        Args:
            instanceId (str): instance identifier '<pdbId>.<authAsymId>'

        Returns:
            bool: True for success or False otherwise
        """
        return instanceId in self.__aD

    def getAssignments(self):
        return self.__aD

    def getAssignmentVersion(self):
        return self.__assignVersion

    def reload(self):
        self.__oD, self.__aD, self.__dumpPath, self.__assignVersion = self.__reload(self.__dirPath, useCache=True)

    def __reload(self, dirPath, **kwargs):
        startTime = time.time()
        oD = {}
        useCache = kwargs.get("useCache", True)
        targetUrl = kwargs.get("targetUrl", "http://opig.stats.ox.ac.uk/webapps/newsabdab/static/downloads/TheraSAbDab_SeqStruc_OnlineDownload.csv")
        #
        ok = False
        fU = FileUtil()
        _, dumpFileName = os.path.split(targetUrl)
        #
        fU.mkdir(dirPath)
        dumpPath = os.path.join(dirPath, dumpFileName)
        dataPath = os.path.join(dirPath, "sabdab-data.json")
        #
        logger.info("useCache %r sabdabDumpPath %r", useCache, dumpPath)
        if useCache and self.__mU.exists(dataPath):
            oD = self.__mU.doImport(dataPath, fmt="json")
        else:
            logger.info("Fetching url %s path %s", targetUrl, dumpPath)
            ok = fU.get(targetUrl, dumpPath)
            #
            rDL = self.__mU.doImport(dumpPath, fmt="csv", rowFormat="dict")
            logger.debug("rD keys %r", list(rDL[0].keys()))
            tD = {}
            for rD in rDL:
                qD = {}
                # Highest_Clin_Trial column name changes each month
                clinTrialCol = next((k for k in list(rD.keys()) if "Highest_Clin_Trial" in k), "Highest_Clin_Trial")
                #
                for kTup in [
                    ("Therapeutic", "antibodyName"),
                    ("Format", "antiBodyFormat"),
                    ("CH1 Isotype", "ch1Isotype"),
                    ("VD LC", "VD_LC"),
                    (clinTrialCol, "maxClinicalPhase"),
                    ("Est. Status", "status"),
                    ("Target", "target"),
                    ("Conditions Approved", "conditionsApproved"),
                    ("Conditions Active", "conditionsActive"),
                ]:
                    if kTup[0] in rD and rD[kTup[0]] not in ["na", "na;na"]:
                        qD[kTup[1]] = rD[kTup[0]]
                    else:
                        qD[kTup[1]] = None
                    if kTup[0] not in rD:
                        logger.error("SabDab key %r missing in input dataset %r", kTup[0], list(rD.keys()))
                tD[rD["Therapeutic"]] = qD
            aD = self.__reloadAssignments(dirPath, **kwargs)
            #
            tS = datetime.datetime.now().isoformat()
            vS = datetime.datetime.now().strftime("%Y-%m-%d")
            oD = {"version": vS, "created": tS, "identifiers": tD, "assignments": aD}
            ok = self.__mU.doExport(dataPath, oD, fmt="json", indent=3)
            logger.info("Exporting (%d) Thera-SAbDab data records and (%d) SAbDab assignments in %r status %r", len(oD["identifiers"]), len(oD["assignments"]), dataPath, ok)

        # ---
        logger.info("Completed reload (%r) at %s (%.4f seconds)", ok, time.strftime("%Y %m %d %H:%M:%S", time.localtime()), time.time() - startTime)
        return oD["identifiers"], oD["assignments"], dumpPath, oD["version"]

    def __reloadAssignments(self, dirPath, **kwargs):
        """Fetch and read

        Args:
            dirPath ([type]): [description]

        Returns:
            [type]: [description]

        """
        startTime = time.time()
        aD = {}
        try:
            targetUrl = kwargs.get("assignmentUrl", "http://opig.stats.ox.ac.uk/webapps/newsabdab/sabdab/summary/all")
            fU = FileUtil()
            dumpFileName = "sabdab_summary_all.tsv"
            #
            fU.mkdir(dirPath)
            dumpPath = os.path.join(dirPath, dumpFileName)
            logger.info("Fetching url %s path %s", targetUrl, dumpPath)
            ok = fU.get(targetUrl, dumpPath)
            rDL = self.__mU.doImport(dumpPath, fmt="tdd", rowFormat="dict")
            logger.info("SAbDab raw records (%d)", len(rDL))
            logger.debug("rD keys %r", list(rDL[0].keys()))
            kyHL = ["pdb", "Hchain", "model", "antigen_chain", "antigen_type", "antigen_het_name", "antigen_name", "heavy_subclass"]
            kyLL = ["pdb", "Lchain", "model", "antigen_chain", "antigen_type", "antigen_het_name", "antigen_name", "light_subclass", "light_ctype"]
            #
            for rD in rDL:
                pdbId = rD["pdb"] if rD["pdb"] and rD["pdb"] != "NA" else None
                authAsymIdH = rD["Hchain"] if rD["Hchain"] and rD["Hchain"] != "NA" else None
                authAsymIdL = rD["Lchain"] if rD["Lchain"] and rD["Lchain"] != "NA" else None
                if pdbId and authAsymIdH:
                    aD[pdbId + "." + authAsymIdH] = {k: v for k, v in rD.items() if v and v != "NA" and k in kyHL}
                if pdbId and authAsymIdL:
                    aD[pdbId + "." + authAsymIdL] = {k: v for k, v in rD.items() if v and v != "NA" and k in kyLL}

            logger.info("Fetched (%d) SAbDab assignment records.", len(aD))
            #
        except Exception as e:
            logger.exception("Failing with %s", str(e))

        logger.info("Completed reload (%r) at %s (%.4f seconds)", ok, time.strftime("%Y %m %d %H:%M:%S", time.localtime()), time.time() - startTime)
        return aD

    def exportFasta(self, fastaPath):
        ok = self.__convertDumpToFasta(dumpPath=self.__dumpPath, fastaPath=fastaPath)
        return ok

    #
    def __convertDumpToFasta(self, dumpPath, fastaPath):
        ok = False
        try:
            rDL = self.__mU.doImport(dumpPath, fmt="csv", rowFormat="dict")
            logger.debug("rD keys %r", list(rDL[0].keys()))
            sD = {}
            for rD in rDL:
                hsKey = "Heavy Sequence" if "Heavy Sequence" in rD else "HeavySequence"
                hSeq = rD[hsKey] if rD[hsKey] != "na" else None
                if not hSeq:
                    continue
                cD = {"sequence": hSeq.strip(), "therapeutic": rD["Therapeutic"], "chain": "heavy"}
                seqId = ""
                cL = []
                for k, v in cD.items():
                    if k in ["sequence"]:
                        continue
                    cL.append(str(v))
                    cL.append(str(k))
                #
                seqId = "|".join(cL)
                sD[seqId] = cD
            for rD in rDL:
                lsKey = "Light Sequence" if "Light Sequence" in rD else "LightSequence"
                lSeq = rD[lsKey] if rD[lsKey] != "na" else None
                if not lSeq:
                    continue
                cD = {"sequence": lSeq.strip(), "therapeutic": rD["Therapeutic"], "chain": "light"}
                seqId = ""
                cL = []
                for k, v in cD.items():
                    if k in ["sequence"]:
                        continue
                    cL.append(str(v))
                    cL.append(str(k))
                #
                seqId = "|".join(cL)
                sD[seqId] = cD
                #
            ok = self.__mU.doExport(fastaPath, sD, fmt="fasta", makeComment=True)
            logger.info("Exporting SAbDab %d fasta sequences in %r status %r", len(sD), fastaPath, ok)
        except Exception as e:
            logger.exception("Failing for %r with %s", dumpPath, str(e))
        return ok



##
#  File:           ChEMBLTargetProvider.py
#  Date:           9-Nov-2020 jdw
#
#  Updated:
#   9-Feb-2023 aae  Update ChEMBL baseVersion to 31
#  18-Jul-2023 dwp  Update ChEMBL baseVersion to 33
##
"""
Accessors for ChEMBL target assignments.

"""

import logging
import os.path
import time

from rcsb.utils.io.FileUtil import FileUtil
from rcsb.utils.io.MarshalUtil import MarshalUtil
from rcsb.utils.seq.UniProtIdMappingProvider import UniProtIdMappingProvider

# pylint: disable=ungrouped-imports
try:
    from chembl_webresource_client.new_client import new_client  # fails when service is down
except Exception:
    pass


logger = logging.getLogger(__name__)


class ChEMBLTargetProvider:
    """Accessors for ChEMBL target assignments."""

    def __init__(self, cachePath, useCache, **kwargs):
        #
        self.__cachePath = cachePath
        self.__dirPath = os.path.join(self.__cachePath, "ChEMBL-targets")
        baseVersion = 34
        self.__version = baseVersion
        self.__mapD = self.__reload(self.__dirPath, baseVersion, useCache, **kwargs)
        #

    def testCache(self, minCount=1):
        return self.__mapD and len(self.__mapD) >= minCount

    def getAssignmentVersion(self):
        return self.__version

    def getTargetDescription(self, unpId):
        try:
            return self.__mapD[unpId][1]
        except Exception:
            logger.error("Missing description for %r", unpId)
            return None

    def getTargetChEMBLId(self, unpId):
        try:
            return self.__mapD[unpId][0]
        except Exception:
            return None

    def getTargetDataPath(self):
        return os.path.join(self.__dirPath, "chembl-target-data.json")

    def reload(self):
        self.__mapD = self.__reload(self.__dirPath, self.__version, useCache=True)

    def __reload(self, dirPath, baseVersion, useCache, **kwargs):
        startTime = time.time()
        mU = MarshalUtil(workPath=dirPath)
        chemblDbUrl = kwargs.get("ChEMBLDbUrl", "https://ftp.ebi.ac.uk/pub/databases/chembl/ChEMBLdb/latest/")
        ok = False
        fU = FileUtil()
        fU.mkdir(dirPath)
        #
        # ChEMBL current version <baseVersion>,...
        # template:  chembl_<baseVersion>.fa.gz
        #
        targetFileName = "chembl_" + str(baseVersion) + ".fa.gz"
        mappingFileName = "chembl_uniprot_mapping.txt"
        #
        chemblTargetPath = os.path.join(dirPath, targetFileName)
        chemblMappingPath = os.path.join(dirPath, mappingFileName)
        mappingFilePath = os.path.join(dirPath, "chembl_uniprot_mapping.json")
        #
        mapD = {}
        if useCache and fU.exists(mappingFilePath):
            logger.info("useCache %r using %r and %r and %r", useCache, chemblTargetPath, chemblMappingPath, mappingFilePath)
            mapD = mU.doImport(mappingFilePath, fmt="json")
        else:
            # Get the ChEMBL UniProt mapping file
            url = os.path.join(chemblDbUrl, mappingFileName)
            ok = fU.get(url, chemblMappingPath)
            logger.info("Fetched %r url %s path %s", ok, url, chemblMappingPath)
            logger.info("Reading ChEMBL mapping file path %s", mappingFilePath)
            rowL = mU.doImport(chemblMappingPath, fmt="tdd", rowFormat="list")
            for row in rowL:
                mapD[row[0]] = (row[1], row[2], row[3])
            ok = mU.doExport(mappingFilePath, mapD, fmt="json")
            logger.info("Processed mapping path %s (%d) %r", mappingFilePath, len(mapD), ok)
            #
            # Get the target FASTA files --
            for vers in range(baseVersion, baseVersion + 10):
                logger.info("Now fetching version %r", vers)
                self.__version = vers
                targetFileName = "chembl_" + str(vers) + ".fa.gz"
                chemblTargetPath = os.path.join(dirPath, "chembl_targets_raw.fa.gz")
                url = os.path.join(chemblDbUrl, targetFileName)
                ok = fU.get(url, chemblTargetPath)
                logger.info("Fetched %r url %s path %s", ok, url, chemblTargetPath)
                if ok:
                    break
        #
        logger.info("Completed reload at %s (%.4f seconds)", time.strftime("%Y %m %d %H:%M:%S", time.localtime()), time.time() - startTime)
        #
        return mapD

    def exportFasta(self, fastaPath, taxonPath, addTaxonomy=False):
        ok = self.__parseFasta(fastaPath, taxonPath, self.__cachePath, self.__dirPath, addTaxonomy=addTaxonomy)
        return ok

    def __parseFasta(self, fastaPath, taxonPath, cachePath, dirPath, addTaxonomy=False):
        # input paths
        chemblTargetRawPath = os.path.join(dirPath, "chembl_targets_raw.fa.gz")
        mU = MarshalUtil(workPath=cachePath)
        oD = {}
        uD = {}
        missTax = 0
        taxonL = []
        try:
            if addTaxonomy:
                umP = UniProtIdMappingProvider(cachePath)
                umP.reload(useCache=True)
            #
            fD = mU.doImport(chemblTargetRawPath, fmt="fasta", commentStyle="default")
            #
            for seqId, sD in fD.items():
                chemblId = seqId.strip().split(" ")[0].strip()
                unpId = seqId[seqId.find("[") + 1 : seqId.find("]")]
                seq = sD["sequence"]
                cD = {"sequence": seq, "uniprotId": unpId, "chemblId": chemblId}
                if addTaxonomy:
                    taxId = umP.getMappedId(unpId, mapName="NCBI-taxon")
                    cD["taxId"] = taxId if taxId else -1
                    if not taxId:
                        missTax += 1
                #
                seqId = ""
                cL = []
                for k, v in cD.items():
                    if k in ["sequence"]:
                        continue
                    cL.append(str(v))
                    cL.append(str(k))
                seqId = "|".join(cL)
                oD[seqId] = cD
                if addTaxonomy:
                    taxonL.append("%s\t%s" % (seqId, taxId))
                #
                uD.setdefault(unpId, []).append(chemblId)
            #
            ok1 = mU.doExport(fastaPath, oD, fmt="fasta", makeComment=True)
            ok3 = True
            if addTaxonomy:
                ok3 = mU.doExport(taxonPath, taxonL, fmt="list")
            return ok1 & ok3
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        #
        return False

    def getActivityData(self, targetChEMBLIdList):
        """Get cofactor activity data for the input ChEMBL target list.

        Args:
            targetChEMBLIdList (list): list of ChEMBL target identifiers

        Returns:
          (dict, dict):  {targetChEMBId: {activity data}}, {moleculeChEMBId: {activity data}}

        """
        targetD = {}
        chunkSize = 50
        numTargets = len(targetChEMBLIdList)
        try:
            for ii in range(0, len(targetChEMBLIdList), chunkSize):
                logger.info("Begin chunk ii %d/%d", ii, numTargets)
                act = new_client.activity  # pylint: disable=no-member
                act.set_format("json")
                actDL = (
                    act.filter(target_chembl_id__in=targetChEMBLIdList[ii : ii + chunkSize]).filter(standard_type__in=["IC50", "Ki", "EC50", "Kd"]).filter(standard_value__isnull=False)
                )
                if actDL:
                    for actD in actDL:
                        targetD.setdefault(actD["target_chembl_id"], []).append(self.__activitySelect(actD))
                logger.info("End chunk completed (%d)", ii)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return targetD

    def __activitySelect(self, aD):
        atL = [
            "assay_chembl_id",
            "assay_description",
            "assay_type",
            "canonical_smiles",
            "ligand_efficiency",
            "molecule_chembl_id",
            "parent_molecule_chembl_id",
            "pchembl_value",
            "standard_relation",
            "standard_type",
            "standard_units",
            "standard_value",
            "target_chembl_id",
        ]
        return {at: aD[at] if at in aD else None for at in atL}

    def getMechanismData(self, targetChEMBLIdList):
        """Get mechanism data for the input ChEMBL target list.

        Args:
            targetChEMBLIdList (list): list of ChEMBL target identifiers

        Returns:
          (dict):  dictionary  {ChEMBId: {mechanism data}}

        """
        oD = {}
        chunkSize = 50
        try:
            for ii in range(0, len(targetChEMBLIdList), chunkSize):
                mch = new_client.mechanism  # pylint: disable=no-member
                mch.set_format("json")
                mDL = mch.filter(target_chembl_id__in=targetChEMBLIdList[ii : ii + chunkSize])
                if mDL:
                    logger.info("mDL (%d)", len(mDL))
                    for mD in mDL:
                        oD.setdefault(mD["target_chembl_id"], []).append(mD)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return oD



##
#  File:           CARDTargetOntologyProvider.py
#  Date:           14-Mar-2023 dwp
#
#  Updates:
#   11-Apr-2023 dwp  Fix issue with lineage tree building--handle cases with two parents at same depth;
#                    Add treeNodeList building and exporting
#   27-Apr-2023 dwp  Update tree node list generation
#    2-May-2023 dwp  Remove depth field from lineage tree
#   27-Aug-2024 dwp  Update reload method to follow behavior of CARDTargetAnnotationProvider,
#                    such that resource data is loaded from stash during cache building workflow
##
"""
Accessors for CARD ontologies.

"""

import datetime
import logging
import os.path
import time

from rcsb.utils.io.FileUtil import FileUtil
from rcsb.utils.io.MarshalUtil import MarshalUtil
from rcsb.utils.io.StashableBase import StashableBase

logger = logging.getLogger(__name__)


class CARDTargetOntologyProvider(StashableBase):
    """Accessors for CARD ontologies."""

    def __init__(self, **kwargs):
        #
        self.__cachePath = kwargs.get("cachePath", ".")
        self.__dirName = "CARD-ontology"
        super(CARDTargetOntologyProvider, self).__init__(self.__cachePath, [self.__dirName])
        self.__dirPath = os.path.join(self.__cachePath, self.__dirName)
        #
        self.__mU = MarshalUtil(workPath=self.__dirPath)
        self.__oD, self.__tnL, self.__version = self.__reload(self.__dirPath, **kwargs)
        #

    def testCache(self, minCount=500):
        ok = False
        if self.__oD and len(self.__oD) > minCount:
            ok = True
        logger.info("CARD ontology testCache status (%r) count %d", ok, len(self.__oD))
        return ok

    def getAssignmentVersion(self):
        return self.__version

    def getLineage(self, aroId):
        """Return the lineage (all parents + the requested aroId) of the given aroId.

        Args:
            aroId (str): ARO ID in the form of "ARO:3001059"

        Returns:
            list: list of dictionaries containing the "id" and "name" of each parent
        """
        return self.__oD.get(aroId, [])

    def getTreeNodeList(self):
        return self.__tnL

    def reload(self):
        self.__oD, self.__tnL, self.__version = self.__reload(self.__dirPath, useCache=True)
        ok = self.testCache()
        return ok

    def __getOntologyDataPath(self, dirPath):
        return os.path.join(dirPath, "card-ontology-data.json")

    def __reload(self, dirPath, **kwargs):
        startTime = time.time()
        oD = {}
        version = None
        tnL = []
        useCache = kwargs.get("useCache", True)
        ok = False
        ontologyDataPath = self.__getOntologyDataPath(dirPath)
        #
        logger.info("useCache %r ontologyDataPath %r", useCache, ontologyDataPath)
        if useCache and self.__mU.exists(ontologyDataPath):
            qD = self.__mU.doImport(ontologyDataPath, fmt="json")
            version = qD["version"]
            oD = qD["data"]
            tnL = qD["treeNodeList"]
            ok = True
        else:
            fU = FileUtil()
            fU.mkdir(dirPath)
        # ---
        logger.info("Completed reload (%r) at %s (%.4f seconds)", ok, time.strftime("%Y %m %d %H:%M:%S", time.localtime()), time.time() - startTime)
        return oD, tnL, version

    def buildOntologyData(self, dirPath=None, **kwargs):
        oD = None
        version = None
        ok = False
        dirPath = dirPath if dirPath else self.__dirPath
        startTime = time.time()
        #
        fU = FileUtil()
        #
        ontologyDumpUrl = kwargs.get("OntologyDumpUrl", "https://card.mcmaster.ca/latest/ontology")
        ontologyDumpFileName = "card-ontology.tar.bz2"
        ontologyDumpPath = os.path.join(dirPath, ontologyDumpFileName)
        ontologyDumpDirPath = os.path.join(dirPath, "dump-ontology")
        #
        fU.mkdir(dirPath)
        ontologyDataPath = self.__getOntologyDataPath(dirPath)
        #
        # ---
        # Build Ontology data
        logger.info("Fetching url %s path %s", ontologyDumpUrl, ontologyDumpPath)
        ok = fU.get(ontologyDumpUrl, ontologyDumpPath)
        fU.mkdir(ontologyDumpDirPath)
        fU.uncompress(ontologyDumpPath, outputDir=ontologyDumpDirPath)
        fU.unbundleTarfile(os.path.join(ontologyDumpDirPath, ontologyDumpFileName[:-4]), dirPath=ontologyDumpDirPath)
        logger.info("Completed fetch (%r) at %s (%.4f seconds)", ok, time.strftime("%Y %m %d %H:%M:%S", time.localtime()), time.time() - startTime)
        oD, tnL, version = self.__parseOntologyData(os.path.join(ontologyDumpDirPath, "aro.obo"))
        #
        tS = datetime.datetime.now().isoformat()
        qD = {"version": version, "created": tS, "data": oD, "treeNodeList": tnL}
        oD = qD["data"]
        ok = self.__mU.doExport(ontologyDataPath, qD, fmt="json", indent=3)
        logger.info("Export CARD ontology data (%d) status %r", len(oD), ok)

        return ok

    def __parseOntologyData(self, filePath):
        """Parse CARD ontology data

        Args:
            filePath (str): card ontology data file (aro.obo)

        Returns:
            (dict, string): dictionary of all ARO IDs and their ancestor lineages, ontology version string
        """
        try:
            cpD, tnL = None, None
            version = None
            parentChildList = []
            idNameD = {}

            with open(filePath, "r", encoding="utf-8") as f:
                data = f.read()

            dL = data.split("[Term]")

            version = dL.pop(0).split("\n")[0].split("format-version: ")[1]

            for tS in dL:
                tsL = tS.strip().split("\n")
                childId, parentId, childName = None, None, None
                for item in tsL:
                    if item.startswith("id: "):
                        childId = item.split('id: ')[1]
                    if item.startswith("is_a: "):
                        parentId = item.split('is_a: ')[1].split(" !")[0]
                    if item.startswith("name: "):
                        childName = item.split('name: ')[1]
                    if parentId and childId and (parentId, childId) not in parentChildList:
                        parentChildList.append((parentId, childId))
                    if childName and childId not in idNameD:
                        idNameD[childId] = childName
                    if item.startswith("[Typedef]"):
                        break

            logger.info("Ontology parent-child pair count (%d)", len(parentChildList))

            cpD, tnL = self.__buildLineageTree(parentChildList, idNameD)

        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return cpD, tnL, version

    def __buildLineageTree(self, parentChildTupleList, idNameMapD):
        """Build a lineage tree containing all children as keys and a list of all possible parents as the values.

        Args:
            parentChildTupleList (list): list of (parent, child) tuples (e.g., [('ARO:1000003', 'ARO:0000000'), ('ARO:1000003', 'ARO:0000001')])
            idNameMapD (dict): dictionary mapping of ARO IDs to their corresponding name

        Returns:
            dict: dictionary containing all children as keys and all possible parents as values
                  (including the child itself, but excluding the top-level parent 'ARO:1000001')
            list: list of all nodes (as dicts) in the tree with their immediate parents only (for building tree in browser)
        """
        # create a dictionary to store the parents of each child
        childToParentD = {}
        for parent, child in parentChildTupleList:
            if child not in childToParentD and child != "ARO:1000001":
                childToParentD[child] = []
            if parent != "ARO:1000001":  # Exclude the top-level "ARO:1000001"
                childToParentD[child].append(parent)

        treeNodeL = self.__exportTreeNodeList(childToParentD, idNameMapD)

        # create a dictionary to store the ancestors of each child
        lineageD = {}
        for child in childToParentD:
            lineageD[child] = [{"id": child, "name": idNameMapD[child]}]  # Add the child to its own ancestry list
            stack = [child]
            while stack:
                node = stack.pop()
                if node in childToParentD:
                    for parent in childToParentD[node]:
                        if parent not in [d["id"] for d in lineageD[child]]:
                            lineageD[child].append({"id": parent, "name": idNameMapD[parent]})
                            stack.append(parent)

        return lineageD, treeNodeL

    def __exportTreeNodeList(self, childToParentD, idNameMapD):
        """Create tree node list in the format of:

        {'id': 'ARO:1000003', 'name': 'antibiotic molecule'}
        {'id': 'ARO:0000041', 'name': 'bacitracin', 'parents': ['ARO:3000053', 'ARO:3000707']}
        {'id': 'ARO:0000039', 'name': 'spectinomycin', 'parents': ['ARO:0000016']}
        """
        #
        dL = []
        for child, parentL in childToParentD.items():
            if parentL:
                tD = {"id": child, "name": idNameMapD[child], "parents": parentL}
            else:
                tD = {"id": child, "name": idNameMapD[child]}
            dL.append(tD)

        return dL



##
#  File:           ChEMBLTargetMechanismProvider.py
#  Date:           9-Nov-2020 jdw
#
#  Updated:
#   9-Feb-2023 aae  Update ChEMBL baseVersion to 31
#  18-Jul-2023 dwp  Update ChEMBL baseVersion to 33
##
"""
Accessors for ChEMBL target mechanism data.

"""

import datetime
import logging
import os.path
import time

from chembl_webresource_client.settings import Settings

# pylint: disable=ungrouped-imports
try:
    from chembl_webresource_client.new_client import new_client  # fails when service is down
except Exception:
    pass

from rcsb.utils.io.FileUtil import FileUtil
from rcsb.utils.io.MarshalUtil import MarshalUtil
from rcsb.utils.io.StashableBase import StashableBase


Settings.Instance().TIMEOUT = 10  # pylint: disable=no-member
Settings.Instance().MAX_LIMIT = 50  # pylint: disable=no-member
Settings.MAX_LIMIT = 50

logger = logging.getLogger(__name__)


class ChEMBLTargetMechanismProvider(StashableBase):
    """Accessors for ChEMBL target mechanism data."""

    def __init__(self, cachePath, useCache):
        #
        self.__cachePath = cachePath
        self.__dirName = "ChEMBL-target-mechanism"
        super(ChEMBLTargetMechanismProvider, self).__init__(self.__cachePath, [self.__dirName])
        self.__dirPath = os.path.join(self.__cachePath, self.__dirName)
        self.__mU = MarshalUtil(workPath=self.__cachePath)
        baseVersion = 33
        self.__version = baseVersion
        logger.info("ChEMBL API MAX_LIMIT %r", Settings.Instance().MAX_LIMIT)  # pylint: disable=no-member
        self.__aD = self.__reload(self.__dirPath, useCache)

    def testCache(self, minCount=1):
        if self.__aD and (len(self.__aD) >= minCount):
            logger.info("Mechanism data for (%d) targets", len(self.__aD))
            return True
        return False

    def getAssignmentVersion(self):
        return self.__version

    def getTargetMechanismDataPath(self):
        return os.path.join(self.__dirPath, "chembl-target-mechanism-data.json")

    def reload(self):
        self.__aD = self.__reload(self.__dirPath, useCache=True)
        return True

    def __reload(self, dirPath, useCache):
        startTime = time.time()
        aD = {}
        fU = FileUtil()
        fU.mkdir(dirPath)
        targetMechanismFilePath = self.getTargetMechanismDataPath()
        #
        if useCache and fU.exists(targetMechanismFilePath):
            logger.info("useCache %r using %r", useCache, targetMechanismFilePath)
            qD = self.__mU.doImport(targetMechanismFilePath, fmt="json")
            aD = qD["mechanism"] if "mechanism" in qD else {}
        #
        logger.info("Completed reload of (%d) at %s (%.4f seconds)", len(aD), time.strftime("%Y %m %d %H:%M:%S", time.localtime()), time.time() - startTime)
        #
        return aD

    def getTargetMechanisms(self, targetChEMBLId):
        try:
            return self.__aD[targetChEMBLId] if targetChEMBLId in self.__aD else []
        except Exception:
            return []

    def hasTargetMechanism(self, targetChEMBLId):
        try:
            return targetChEMBLId in self.__aD
        except Exception:
            return False

    def fetchTargetMechanismData(self, targetChEMBLIdList, skipExisting=True, chunkSize=50):
        """Get cofactor mechanism data for the input ChEMBL target list.

        Args:
            targetChEMBLIdList (list): list of ChEMBL target identifiers
            skipExisting (bool, optional): reuse any existing cached data (default: True)
            chunkSize(int, optional): ChEMBL API batch size for fetches (default: 50)

        Returns:
          bool:  True for success or False otherwise

        """
        atL = [
            "action_type",
            "molecule_chembl_id",
            "action_type",
            "mechanism_of_action",
            "max_phase",
            "target_chembl_id",
        ]
        targetD = self.__aD if self.__aD else {}
        idList = []
        if skipExisting:
            for tId in targetChEMBLIdList:
                if tId in self.__aD:
                    continue
                idList.append(tId)
        else:
            idList = targetChEMBLIdList

        numToProcess = len(idList)
        logger.info("Fetching mechanism data for (%d/%d)", numToProcess, len(targetChEMBLIdList))
        ok = False
        try:
            for ii in range(0, len(idList), chunkSize):
                logger.info("Begin chunk at ii %d/%d", ii, numToProcess)
                mch = new_client.mechanism  # pylint: disable=no-member
                mch.set_format("json")
                mDL = mch.filter(target_chembl_id__in=idList[ii : ii + chunkSize]).only(atL)

                logger.info("Results (%d)", len(mDL))
                if mDL:
                    for mD in mDL:
                        targetD.setdefault(mD["target_chembl_id"], []).append(self.__mechanismSelect(atL, mD))
                #
                logger.info("Completed chunk starting at (%d)", ii)
                tS = datetime.datetime.now().isoformat()
                vS = datetime.datetime.now().strftime("%Y-%m-%d")
                ok = self.__mU.doExport(self.getTargetMechanismDataPath(), {"version": vS, "created": tS, "mechanism": targetD}, fmt="json", indent=3)
                logger.info("Wrote completed chunk starting at (%d) (%r)", ii, ok)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return ok

    def __mechanismSelect(self, atL, aD):
        return {at: aD[at] if at in aD else None for at in atL}



##
#  File:           ChEMBLTargetCofactorProvider.py
#  Date:           15-Jun-2021 jdw
#
#  Updated:
#  20-Aug-2024 dwp Add support for loading and accessing data on MongoDB
##
"""
Accessors for ChEMBL target cofactors.
"""

import datetime
import logging
import math
import os.path
import time

from rcsb.utils.io.FileUtil import FileUtil
from rcsb.utils.io.MarshalUtil import MarshalUtil
from rcsb.utils.io.StashableBase import StashableBase
from rcsb.utils.targets.ChEMBLTargetActivityProvider import ChEMBLTargetActivityProvider
from rcsb.utils.targets.ChEMBLTargetProvider import ChEMBLTargetProvider
from rcsb.utils.targets.TargetCofactorDbProvider import TargetCofactorDbProvider

logger = logging.getLogger(__name__)


class ChEMBLTargetCofactorProvider(StashableBase):
    """Accessors for ChEMBL target cofactors."""

    def __init__(self, **kwargs):
        #
        self.__cachePath = kwargs.get("cachePath", ".")
        self.__dirName = "ChEMBL-cofactors"
        self.__cofactorResourceName = "chembl"
        super(ChEMBLTargetCofactorProvider, self).__init__(self.__cachePath, [self.__dirName])
        self.__dirPath = os.path.join(self.__cachePath, self.__dirName)
        #
        self.__mU = MarshalUtil(workPath=self.__dirPath)
        self.__fD = self.__reload(self.__dirPath, **kwargs)
        #

    def testCache(self, minCount=1):
        logger.info("ChEMBL cofactor count %d", len(self.__fD["cofactors"]) if "cofactors" in self.__fD else 0)
        if self.__fD and "cofactors" in self.__fD and len(self.__fD["cofactors"]) > minCount:
            return True
        else:
            return False

    def hasTarget(self, rcsbEntityId):
        return rcsbEntityId.upper() in self.__fD["cofactors"]

    def getTargets(self, rcsbEntityId):
        try:
            return self.__fD["cofactors"][rcsbEntityId.upper()]
        except Exception:
            return []

    def __getCofactorDataPath(self):
        return os.path.join(self.__dirPath, "ChEMBL-cofactor-data.json")

    def reload(self):
        self.__fD = self.__reload(self.__dirPath, useCache=True)
        return True

    def __reload(self, dirPath, **kwargs):
        startTime = time.time()
        fD = {}
        useCache = kwargs.get("useCache", True)
        ok = False
        cofactorPath = self.__getCofactorDataPath()
        #
        logger.info("useCache %r cofactorPath %r", useCache, cofactorPath)
        if useCache and self.__mU.exists(cofactorPath):
            fD = self.__mU.doImport(cofactorPath, fmt="json")
            ok = len(fD) > 0
        else:
            fU = FileUtil()
            fU.mkdir(dirPath)
        # ---
        logger.info("Completed reload with status (%r) at %s (%.4f seconds)", ok, time.strftime("%Y %m %d %H:%M:%S", time.localtime()), time.time() - startTime)
        return fD

    def getCofactorDataDict(self):
        return self.__fD["cofactors"]

    def buildCofactorList(self, sequenceMatchFilePath, crmpObj=None, lnmpObj=None, maxActivity=5):
        """Build target cofactor list for the matching entities in the input sequence match file.

        Args:
            sequenceMatchFilePath (str): sequence match output file path (e.g., "./CACHE/sequence-search-results/chembl-vs-pdbprent-filtered-results.json")
            crmpObj (obj, optional): instance of ChemRefMappingProviderObj()
            lnmpObj (obj, optional): instance of LigandNeighborMappingProviderObj(). Defaults to None.
            maxActivity (int, optional): maximum number of prioritized activity records per target

        Returns:
            bool: True for success or False otherwise

            Example activity record -

                        "CHEMBL3243": [
                    {
                        "assay_chembl_id": "CHEMBL655768",
                        "assay_description": "In vitro inhibitory activity against recombinant human CD45 using fluorescein diphosphate (FDP) as a substrate",
                        "assay_type": "B",
                        "canonical_smiles": "COC(=O)c1ccc(C2=CC(=O)C(=O)c3ccccc32)cc1",
                        "ligand_efficiency": {
                        "bei": "19.78",
                        "le": "0.36",
                        "lle": "3.11",
                        "sei": "9.57"
                        },
                        "molecule_chembl_id": "CHEMBL301254",
                        "parent_molecule_chembl_id": "CHEMBL301254",
                        "pchembl_value": "5.78",
                        "standard_relation": "=",
                        "standard_type": "IC50",
                        "standard_units": "nM",
                        "standard_value": "1650.0",
                        "target_chembl_id": "CHEMBL3243"
                    },
        """
        rDL = []
        mD = self.__mU.doImport(sequenceMatchFilePath, fmt="json")
        #
        chP = ChEMBLTargetProvider(cachePath=self.__cachePath, useCache=False)
        if not chP.testCache():
            logger.warning("Skipping build of target cofactor list because ChEMBL Target data is missing.")
            return False
        # ---
        chaP = ChEMBLTargetActivityProvider(cachePath=self.__cachePath, useCache=True)
        if not chaP.testCache():
            logger.warning("Skipping build of target cofactor list because ChEMBL Target Activity data is missing.")
            return False
        #
        provenanceSource = "ChEMBL"
        refScheme = "PDB entity"
        assignVersion = chP.getAssignmentVersion()
        for queryId, matchDL in mD.items():
            qCmtD = self.__decodeComment(queryId)
            unpId = qCmtD["uniprotId"]
            queryTaxId = qCmtD["taxId"] if "taxId" in qCmtD else None
            chemblIdL = qCmtD["chemblId"].split(",")
            if queryTaxId == "-1":
                logger.info("Skipping target with missing taxonomy %r (%r)", unpId, chemblIdL)
                continue
            queryName = chP.getTargetDescription(unpId)
            for chemblId in chemblIdL:
                if not chaP.hasTargetActivity(chemblId):
                    logger.debug("Skipping target %r (%r)", unpId, chemblId)
                    # continue
                # --
                chemCompNeighborsD = {}
                if lnmpObj:
                    for matchD in matchDL:
                        tCmtD = self.__decodeComment(matchD["target"])
                        entryId = tCmtD["entityId"].split("_")[0]
                        entityId = tCmtD["entityId"].split("_")[1]
                        rcsbEntityId = entryId + "_" + entityId
                        chemCompIdList = lnmpObj.getLigandNeighbors(rcsbEntityId)
                        chemCompNeighborsD.update({k: True for k in chemCompIdList})
                # --
                for matchD in matchDL:
                    tCmtD = self.__decodeComment(matchD["target"])
                    entryId = tCmtD["entityId"].split("_")[0]
                    entityId = tCmtD["entityId"].split("_")[1]
                    #
                    taDL = chaP.getTargetActivity(chemblId)
                    logger.debug("Target %r has (%d) activity records", chemblId, len(taDL))
                    # ---
                    actL = []
                    for taD in taDL:
                        if taD["assay_type"] in ["B", "F"]:
                            try:
                                if taD["standard_units"] == "nM" and taD["standard_value"] and float(taD["standard_value"]) > 0.0:
                                    pV = -math.log10(float(taD["standard_value"]) * 10.0e-9)
                                    actD = {
                                        "cofactor_id": taD["molecule_chembl_id"],
                                        "assay_id": taD["assay_chembl_id"],
                                        "assay_description": taD["assay_description"],
                                        "measurement_type": "p" + taD["standard_type"],
                                        "measurement_value": round(pV, 2),
                                        "smiles": taD["canonical_smiles"],
                                        "molecule_name": taD["molecule_name"],
                                        "inchi_key": taD["inchi_key"],
                                        "action": taD["action"],
                                        "moa": taD["moa"],
                                        "max_phase": taD["max_phase"],
                                    }
                                    actD = self.__addLocalIds(actD, crmpObj=crmpObj)
                                    actL.append(actD)
                            except Exception as e:
                                logger.debug("Failing for tAD %r with %s", taD, str(e))

                    # ---
                    actL = self.__activityListSelect(actL, chemCompNeighborsD, maxActivity=maxActivity)
                    if not actL:
                        logger.debug("No ChEMBL cofactors for %s %s", chemblId, unpId)
                    # ---
                    # aligned_target.entity_beg_seq_id (current target is PDB entity in json)
                    # aligned_target.target_beg_seq_id (current query is target seq in json)
                    # aligned_target.length
                    fpL = []
                    if "alignedRegions" in matchD:
                        fpL = [
                            {
                                "entity_beg_seq_id": arD["targetBegin"],
                                "target_beg_seq_id": arD["queryBegin"],
                                "length": arD["targetEnd"] - arD["targetBegin"],
                            }
                            for arD in matchD["alignedRegions"]
                        ]
                    else:
                        fpL = [
                            {
                                "entity_beg_seq_id": matchD["targetBegin"],
                                "target_beg_seq_id": matchD["queryBegin"],
                                "length": matchD["alignLen"],
                            }
                        ]
                    # ---
                    rD = {
                        "entry_id": entryId,
                        "entity_id": entityId,
                        "query_uniprot_id": unpId,
                        "query_id": chemblId,
                        "query_id_type": "ChEMBL",
                        "query_name": queryName,
                        "provenance_source": provenanceSource,
                        "reference_scheme": refScheme,
                        "assignment_version": assignVersion,
                        "query_taxonomy_id": int(queryTaxId) if queryTaxId else None,
                        "target_taxonomy_id": int(matchD["targetTaxId"]) if "targetTaxId" in matchD else None,
                        #
                        "aligned_target": fpL,
                        #
                        "taxonomy_match_status": matchD["taxonomyMatchStatus"] if "taxonomyMatchStatus" in matchD else None,
                        "lca_taxonomy_id": matchD["lcaTaxId"] if "lcaTaxId" in matchD else None,
                        "lca_taxonomy_name": matchD["lcaTaxName"] if "lcaTaxName" in matchD else None,
                        "lca_taxonomy_rank": matchD["lcaRank"] if "lcaRank" in matchD else None,
                        "cofactors": actL,
                    }
                    rDL.append(rD)
            #
        qD = {}
        for rD in rDL:
            eId = rD["entry_id"] + "_" + rD["entity_id"]
            qD.setdefault(eId, []).append(rD)
        #
        tS = datetime.datetime.now().isoformat()
        # vS = datetime.datetime.now().strftime("%Y-%m-%d")
        vS = assignVersion
        #
        # Write out cofactor data set
        fp = self.__getCofactorDataPath()
        ok = self.__mU.doExport(fp, {"version": vS, "created": tS, "cofactors": qD}, fmt="json", indent=3)
        #
        return ok

    def __addLocalIds(self, cfD, crmpObj=None):
        #
        if crmpObj:
            localIdL = crmpObj.getLocalIds("CHEMBL", cfD["cofactor_id"])
            if localIdL:
                localId = localIdL[0]
                if localId.startswith("PRD_"):
                    cfD["prd_id"] = localId
                else:
                    cfD["chem_comp_id"] = localId
        return cfD

    def __activityListSelect(self, activityDL, chemCompNeighborsD, maxActivity=5):
        retL = []
        mappedNeighborL = []
        unmappedL = activityDL
        #
        if chemCompNeighborsD:
            unmappedL = []
            # Select out the any cases for molecules that map to a neighbor chemical component.
            for activityD in activityDL:
                if "chem_comp_id" in activityD and activityD["chem_comp_id"] in chemCompNeighborsD:
                    activityD["neighbor_in_pdb"] = "Y"
                    mappedNeighborL.append(activityD)
                else:
                    unmappedL.append(activityD)
                    activityD["neighbor_in_pdb"] = "N"
        #
        numLeft = maxActivity - len(mappedNeighborL)
        if numLeft > 0:
            unmappedL = sorted(unmappedL, key=lambda k: k["measurement_value"], reverse=True)
            retL = mappedNeighborL
            retL.extend(unmappedL[:numLeft])
            retL = sorted(retL, key=lambda k: k["measurement_value"], reverse=True)
        else:
            logger.debug("Mapped neighbor cofactors (%d) excluded unmapped (%d)", len(mappedNeighborL), len(unmappedL))
            retL = sorted(mappedNeighborL, key=lambda k: k["measurement_value"], reverse=True)
        return retL

    def __decodeComment(self, comment, separator="|"):
        dD = {}
        try:
            ti = iter(comment.split(separator))
            dD = {tup[1]: tup[0] for tup in zip(ti, ti)}
        except Exception:
            pass
        return dD

    def loadCofactorData(self, cfgOb, **kwargs):
        """Load cofactor data to MongoDB.
        """
        ok = False
        try:
            tcDbP = TargetCofactorDbProvider(
                cachePath=self.__cachePath,
                cfgOb=cfgOb,
                cofactorResourceName=self.__cofactorResourceName,
                **kwargs
            )
            ok = tcDbP.loadCofactorData(self.__cofactorResourceName, self)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        #
        logger.info("%r cofactor data DB load status (%r)", self.__cofactorResourceName, ok)
        return ok


class ChEMBLTargetCofactorAccessor:
    def __init__(self, cachePath, cfgOb=None, **kwargs):
        """
        Accessor class for fetching cofactor data from MongoDB.
        """
        self.__cofactorResourceName = "chembl"
        self.__cfgOb = cfgOb
        self.__cachePath = cachePath
        #
        self.__tcDbP = TargetCofactorDbProvider(
            cachePath=self.__cachePath,
            cfgOb=self.__cfgOb,
            cofactorResourceName=self.__cofactorResourceName,
            **kwargs
        )

    def testCache(self, minCount=0):
        docCount = self.__tcDbP.cofactorDbCount()
        logger.info("Loaded %s cofactor DB count %d", self.__cofactorResourceName, docCount)
        return docCount >= minCount

    def reload(self):
        return True

    def getTargets(self, rcsbEntityId, dataFieldName="rcsb_cofactors", **kwargs):
        return self.__tcDbP.fetchCofactorData(rcsbEntityId, dataFieldName, **kwargs)
